/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - linuxMib2.c */
/* rdk@krupczak.org */

/* a plugin implementing MIB-2 objects for Linux */

#ifdef LINUX

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <net/if_arp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#if 0

/* depending on the version of linux, we need to include 
   linux/tcp.h or netinet/tcp.h but often not both */

#include <linux/tcp.h>
#include <linux/udp.h>
#endif /* 0 */

#include <sys/ioctl.h>
#include <net/route.h>  /* route flags */

#include "xmp.h"
#include "mib2plugin.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "pluginexport.h"
#include "korgdebug.h"

static pthread_t wt;
extern int mib2ThreadKill;

mib2TcpConn_t *tcpTable = NULL;
int numTcpConnections = 0, maxTcpConnections = 0;
time_t tcpTableTime = 0;

mib2UdpEndpt_t *udpTable = NULL;
int numUdps = 0, maxUdps = 0;
time_t udpTableTime = 0;

mib2IfEntry_t *ifTable = NULL;
int numIfs = 0, maxIfs = 0;
time_t ifTableTime = 0;

ipRoute_t *routeTable = NULL;
int numRoutes = 0, maxRoutes = 0;
time_t routeTableTime = 0;

arpEntry_t *arpTable = NULL;
int numArps = 0, maxArps = 0;
time_t arpTableTime = 0;

ipAddrEntry_t *addrTable = NULL;
int numAddrs = 0, maxAddrs = 0;
time_t ipAddrTableTime = 0;

time_t statsTime = 0;
ipStats_t ipstats;
tcpStats_t tcpstats;
icmpStats_t icmpstats;
udpStats_t udpstats;

int debugLevel;

pthread_mutex_t mib2_lock = PTHREAD_MUTEX_INITIALIZER;

/* need to ensure that addrs are in network-byte order, all addresses */

/* ************************************************************* */
/* convert a 32-char string of hex digits into an RFC 2373 address
   format string; return length of converted string used for Linux
   /proc/net/tcp6 and /proc/net/udp6 grog'ing 

   in /proc/net/blah6 files, local addresses come out as
   4 32-bit words each in network byte order; consequently,
   we need to parse and order them properly

*/

int convertIPv6ByteStr(char *str,char *converted,int maxsize)
{
     u_int a1,a2,a3,a4;
     u_char *byteStr;
     char *convertedPtr;

     if ((str == NULL) || (converted == NULL)) {
        return -1;
     }
     if (maxsize < 48) {
        return -1;
     }

     sscanf(str,"%08x%08x%08x%08x",&a1,&a2,&a3,&a4);
     convertedPtr = converted;
 
     /* take 4-char chunks at a time */
     byteStr = (u_char *)&a1;

     sprintf(convertedPtr,"%02x",byteStr[0]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[1]); convertedPtr += 2;
     sprintf(convertedPtr,":"); convertedPtr += 1;
     sprintf(convertedPtr,"%02x",byteStr[2]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[3]); convertedPtr += 2;

     sprintf(convertedPtr,":"); convertedPtr += 1;
     byteStr = (u_char *)&a2;

     sprintf(convertedPtr,"%02x",byteStr[0]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[1]); convertedPtr += 2;
     sprintf(convertedPtr,":"); convertedPtr += 1;
     sprintf(convertedPtr,"%02x",byteStr[2]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[3]); convertedPtr += 2;

     sprintf(convertedPtr,":"); convertedPtr += 1;
     byteStr = (u_char *)&a3; 

     sprintf(convertedPtr,"%02x",byteStr[0]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[1]); convertedPtr += 2;
     sprintf(convertedPtr,":"); convertedPtr += 1;
     sprintf(convertedPtr,"%02x",byteStr[2]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[3]); convertedPtr += 2;

     sprintf(convertedPtr,":"); convertedPtr += 1;
     byteStr = (u_char *)&a4;

     sprintf(convertedPtr,"%02x",byteStr[0]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[1]); convertedPtr += 2;
     sprintf(convertedPtr,":"); convertedPtr += 1;
     sprintf(convertedPtr,"%02x",byteStr[2]); convertedPtr += 2;
     sprintf(convertedPtr,"%02x",byteStr[3]); convertedPtr += 2;

     *convertedPtr = 0; 

     DBG1("convertIPv6ByteStr: converted str is %s\n",converted);

     return strlen(converted);
}

/* ************************************************************* */
/* caller should already have lock for table */

static int enlarge_udp_table()
{
     /* if needing initialization */
     if (udpTable == NULL) {
        udpTable = (mib2UdpEndpt_t *)malloc(sizeof(mib2UdpEndpt_t) * 50);
        if (udpTable == NULL) {
	   perror("enlarge_udp_table: initial malloc failed");
           return -1;
        }
        maxUdps = 50;
     }

     if (maxUdps == numUdps) {
        maxUdps += 25;
        udpTable = (mib2UdpEndpt_t *)realloc(udpTable,
		    maxUdps*sizeof(mib2UdpEndpt_t));

        if (udpTable == NULL) {
	   perror("enlarge_udp_table: realloc failed");
           maxUdps = numUdps = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */
/* caller should already have lock for table */

static int enlarge_tcp_table()
{
     /* if needing initialization */
     if (tcpTable == NULL) {
        tcpTable = (mib2TcpConn_t *)malloc(sizeof(mib2TcpConn_t) * 50);
        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: initial malloc failed");
           return -1;
        }
        maxTcpConnections = 50;
     }

     if (maxTcpConnections == numTcpConnections) {
        maxTcpConnections += 25;
        tcpTable = (mib2TcpConn_t *)realloc(tcpTable,
		    maxTcpConnections*sizeof(mib2TcpConn_t));

        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: realloc failed");
           maxTcpConnections = numTcpConnections = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_if_table()
{
     /* if needing initialization */
     if (ifTable == NULL) {
        ifTable = (mib2IfEntry_t *)malloc(sizeof(mib2IfEntry_t) * 5);
        if (ifTable == NULL) {
	   perror("enlarge_if_table: initial malloc failed");
           return -1;
        }
        maxIfs = 5;
     }

     if (maxIfs == numIfs) {
        maxIfs += 5;
        ifTable = (mib2IfEntry_t *)realloc(ifTable,
		    maxIfs*sizeof(mib2IfEntry_t));

        if (ifTable == NULL) {
	   perror("enlarge_if_table: realloc failed");
           maxIfs = numIfs = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_route_table()
{
     /* if needing initialization */
     if (routeTable == NULL) {
        routeTable = (ipRoute_t *)malloc(sizeof(ipRoute_t) * 5);
        if (routeTable == NULL) {
	   perror("enlarge_route_table: initial malloc failed");
           return -1;
        }
        maxRoutes = 5;
     }

     if (maxRoutes == numRoutes) {
        maxRoutes += 5;
        routeTable = (ipRoute_t *)realloc(routeTable,
		    maxRoutes*sizeof(ipRoute_t));

        if (routeTable == NULL) {
	   perror("enlarge_route_table: realloc failed");
           maxRoutes = numRoutes = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_arp_table()
{
     /* if needing initialization */
     if (arpTable == NULL) {
        arpTable = (arpEntry_t *)malloc(sizeof(arpEntry_t) * 5);
        if (arpTable == NULL) {
	   perror("enlarge_arp_table: initial malloc failed");
           return -1;
        }
        maxArps = 5;
     }

     if (maxArps == numArps) {
        maxArps += 5;
        arpTable = (arpEntry_t *)realloc(arpTable,
		    maxArps*sizeof(arpEntry_t));

        if (arpTable == NULL) {
	   perror("enlarge_arp_table: realloc failed");
           maxArps = numArps = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_addr_table()
{
     /* if needing initialization */
     if (addrTable == NULL) {
        addrTable = (ipAddrEntry_t *)malloc(sizeof(ipAddrEntry_t) * 5);
        if (addrTable == NULL) {
	   perror("enlarge_addr_table: initial malloc failed");
           return -1;
        }
        maxAddrs = 5;
     }

     if (maxAddrs == numAddrs) {
        maxAddrs += 5;
        addrTable = (ipAddrEntry_t *)realloc(addrTable,
		    maxAddrs*sizeof(ipAddrEntry_t));

        if (addrTable == NULL) {
	   perror("enlarge_addr_table: realloc failed");
           maxAddrs = numAddrs = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

u_char convertTCPState(u_char sockState)
{
     switch (sockState) {
       case TCP_ESTABLISHED:
	  return tcpStateEstablished;
          break;
       case TCP_SYN_SENT:
	  return tcpStateSynSent;
          break;
       case TCP_SYN_RECV:
	  return tcpStateSynReceived;
          break;
       case TCP_FIN_WAIT1:
	  return tcpStateFinWait1;
          break;
       case TCP_FIN_WAIT2:
	  return tcpStateFinWait2;
          break;
       case TCP_TIME_WAIT:
	  return tcpStateTimeWait;
          break;
       case TCP_CLOSE:
	  return tcpStateClosed;
          break;
       case TCP_CLOSE_WAIT:
	  return tcpStateCloseWait;
          break;
       case TCP_LAST_ACK:
	  return tcpStateLastAck;
          break;
       case TCP_LISTEN:
	  return tcpStateListen;
          break;
       case TCP_CLOSING:
	  return tcpStateClosing;
          break;
       default:
          return tcpStateUndef;
     }

     return tcpStateUndef;
}

/* in udp.c, kernel is writing state of sock structure
   not socket struct 
   it appears as if tcp states CLOSE and ESTABLISHED
   CLOSE implies unconnected
   ESTABLISHED implies connected socket 
*/

u_char convertUDPState(u_char sockState)
{
     switch (sockState) {
       case TCP_ESTABLISHED:
	  return udpStateConnected;
          break;

       case TCP_CLOSE:
	  return udpStateUnconnected;
          break;

       case TCP_SYN_SENT:
       case TCP_SYN_RECV:
       case TCP_FIN_WAIT1:
       case TCP_FIN_WAIT2:
       case TCP_TIME_WAIT:
       case TCP_CLOSE_WAIT:
       case TCP_LAST_ACK:
       case TCP_LISTEN:
       case TCP_CLOSING:
       default:
	  return udpStateUnknown;
          break;
     }
}

/* ************************************************************* */
/* get tcp connection table from /proc/net/tcp only IPv4 TCP connections; 

   IP addresses coming out of his /proc/net/tcp file are in 
   network byte order and thus need to be converted prior to 
   ascii manipulation ; ports seem to come out in host byte order though
   Confirmed these assertions with testproc.c test program

   on x86, host byte order is LSB first
  
   Network byte order is MSB first

   state/status comes from tcp.h 
 */

static void reloadTcpTable()
{
     FILE *fp;
     char line[4096];
     int i;
     char *aPtr, aToken[256], *lasts, ipv6addr[48];
     int inodeCol = 0;
     int localCol = 0;
     int remCol = 0;
     int stateCol = 0;
     u_int sockState;
     char ipstr1[64],ipstr2[64];

     /* get lock on tcpTable */
     pthread_mutex_lock(&mib2_lock);

     if ((fp = fopen("/proc/net/tcp","r")) == NULL) {
        perror("reloadTcpTable: fopen tcp failed");
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* read first line; it should be column headers; find out
        which columns are local and remote addresses */
     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadTcpTable: empty tcp file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }
     i = 0;
     aPtr = line;
     aPtr = strtok_r(aPtr," \t\n",&lasts);
     while (aPtr != NULL) {

       if (strcasecmp(aPtr,"local_address") == 0) {
          localCol = i;
       }
       if (strcasecmp(aPtr,"rem_address") == 0) {
          remCol = i;
       }
       if (strcasecmp(aPtr,"st") == 0) {
 	  stateCol = i;
       }
       if (strcasecmp(aPtr,"inode") == 0) {
	  inodeCol = i-2; /* magic -2 */

          /* magic -2 because the header in /proc/net/tcp and /proc/net/udp
             is sort of aligned by column char position and not
             necessarily indicative of column number; consequently
             tx_queue/rx_queue are separate columns in the
             header-line but not separated by ' ' in the actual
             data-line.  Same for tr and tr->when 
           */

       }

       aPtr = strtok_r(NULL," \t\n",&lasts);
       
       i++;
     }

     if (debugLevel)
        printf("reloadTcpTable: local col:%d rem col:%d\n",
               localCol,remCol);

     /* if we couldnt find both columns, give up */
     if ((localCol == 0) || (remCol == 0) || (stateCol == 0)) {
        printf("reloadTcpTable: unable to discern cols %d, %d\n",
                localCol,remCol);
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pre-allocate some space if we need to */
     enlarge_tcp_table();

     /* now, loop through line by line examining tcp connections */
     numTcpConnections = 0;
     while (fgets(line,sizeof(line)-1,fp) != NULL) {

       tcpTable[numTcpConnections].localIPv6[0] = 0;
       tcpTable[numTcpConnections].remIPv6[0] = 0;
       tcpTable[numTcpConnections].ipVersion = 4;

       i=0;
       aPtr = line;
       aPtr = strtok_r(aPtr," \t\n",&lasts);
       while (aPtr != NULL) {
         if (i == localCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%08x:%04x",
                  &tcpTable[numTcpConnections].localIp,
                  &tcpTable[numTcpConnections].localPort);
         }
         if (i == remCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%08x:%04x",
                  &tcpTable[numTcpConnections].remIp,
                  &tcpTable[numTcpConnections].remPort);
         }
         if (i == stateCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%x",&sockState);
            tcpTable[numTcpConnections].state = (u_char)sockState;
            tcpTable[numTcpConnections].state = 
	      convertTCPState(sockState);
         }
         if (i == inodeCol) {
	    strcpy(aToken,aPtr);
            sscanf(aToken,"%llu",&tcpTable[numTcpConnections].inode);
         }

         aPtr = strtok_r(NULL," \t\n",&lasts);
         i++;

       } /* while strtok through the line */


   xmpConvertIpAddr(tcpTable[numTcpConnections].localIp,ipstr1,sizeof(ipstr1));
   xmpConvertIpAddr(tcpTable[numTcpConnections].remIp,ipstr2,sizeof(ipstr2));

     if (debugLevel)
        printf("reloadTcpTable: State:%02x %s:%d %s:%d, inode:%llu\n",
              sockState,ipstr1,
              tcpTable[numTcpConnections].localPort,
              ipstr2,
              tcpTable[numTcpConnections].remPort,
              tcpTable[numTcpConnections].inode);

       numTcpConnections++;
       enlarge_tcp_table(); /* enlarge if we need more space */

     } /* fgets line */

     fclose(fp);

     /* ********************************************************* */
     /* now, open/parse /proc/net/tcp6 since many linux/linux
        connections are now defaulting over tcp6 */
     if ((fp = fopen("/proc/net/tcp6","r")) == NULL) {
        ; /* most likely error is tcp6 does not exist */
     }
     else {
        /* read first line; it should be column headers; find out
           which columns are local and remote addresses */
        localCol = remCol = stateCol = inodeCol = 0;
        if (fgets(line,sizeof(line)-1,fp) == NULL) {
           perror("reloadTcpTable: empty tcp6 file");
        }
        else {
           i = 0;
           aPtr = line;
           aPtr = strtok_r(aPtr," \t\n",&lasts);
           while (aPtr != NULL) {
             if (strcasecmp(aPtr,"local_address") == 0) {
                localCol = i;
             }
             if (strcasecmp(aPtr,"remote_address") == 0) {
                remCol = i;
             }
             if (strcasecmp(aPtr,"st") == 0) {
                stateCol = i;
             }
             if (strcasecmp(aPtr,"inode") == 0) {
	        inodeCol = i-2; /* magic -2 */
             }
             aPtr = strtok_r(NULL," \t\n",&lasts);
             i++;
           } /* fgets first line */

           /* now get each entry */

           while (fgets(line,sizeof(line)-1,fp) != NULL) {

             tcpTable[numTcpConnections].ipVersion = 6;
             tcpTable[numTcpConnections].localIp = 0;
             tcpTable[numTcpConnections].remIp = 0;
             i = 0;
             aPtr = line;
             aPtr = strtok_r(aPtr," \t\n",&lasts);
             while (aPtr != NULL) {
               if (i == localCol) {
                  strcpy(aToken,aPtr);
                  sscanf(aToken,"%32s:%04x",
			ipv6addr,
                        &tcpTable[numTcpConnections].localPort);
                  /* convert the address */
                  convertIPv6ByteStr(ipv6addr,
                                     tcpTable[numTcpConnections].localIPv6,
                                     48);
               }
               if (i == remCol) {
 	          strcpy(aToken,aPtr);
                  sscanf(aToken,"%32s:%04x",
			 ipv6addr,
                         &tcpTable[numTcpConnections].remPort);
                  convertIPv6ByteStr(ipv6addr,
                                     tcpTable[numTcpConnections].remIPv6,
                                     48);
               }
               if (i == stateCol) {
 	          strcpy(aToken,aPtr);
                  sscanf(aToken,"%x",&sockState);
                  tcpTable[numTcpConnections].state = (u_char)sockState;
                  tcpTable[numTcpConnections].state = 
                          convertTCPState(sockState);
               }
               if (i == inodeCol) {
	          strcpy(aToken,aPtr);
                  sscanf(aToken,"%llu",&tcpTable[numTcpConnections].inode);
               }
               aPtr = strtok_r(NULL," \t\n",&lasts);
               i++;

            } /* while strtok through the line */

            numTcpConnections++;
            enlarge_tcp_table(); /* enlarge if we need more space */

           } /* while fgets line */

	} /* else read first line */

        fclose(fp);
     } /* open tcp6 */

     /* sort? */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadTcpTable: %d connections found\n",numTcpConnections);

     tcpTableTime = time(NULL);
}

/* ************************************************************* */
/* get UDP endpoints from /proc/net/udp
   only IPv4 UDP endpoints
   IP addresses coming out of this file are in network-byte
   and thus need to be converted prior to ascii manipulation
   ports seem to come out in host byteorder though XXX
   state/status comes from udp.h ?
*/

static void reloadUdpTable()
{
     FILE *fp;
     char line[4096];
     int i;
     char *aPtr, aToken[256], *lasts, ipv6addr[48];
     int inodeCol = 0;
     int localCol = 0;
     int remCol = 0;
     int stateCol = 0;
     u_int sockState;
     char ipstr1[64],ipstr2[64];

     /* get lock on udpTable */
     pthread_mutex_lock(&mib2_lock);

     if ((fp = fopen("/proc/net/udp","r")) == NULL) {
        perror("reloadUdpTable: fopen udp failed");
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* read first line; it should be column headers; find out
        which columns are local and remote addresses */

     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadUdpTable: empty udp file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }
     i = 0;
     aPtr = line;
     aPtr = strtok_r(aPtr," \t\n",&lasts);
     while (aPtr != NULL) {

       if (strcasecmp(aPtr,"local_address") == 0) {
          localCol = i;
       }
       if (strcasecmp(aPtr,"rem_address") == 0) {
          remCol = i;
       }
       if (strcasecmp(aPtr,"st") == 0) {
 	  stateCol = i;
       }
       if (strcasecmp(aPtr,"inode") == 0) {
	  inodeCol = i-2; /* magic -2 */
       }

       aPtr = strtok_r(NULL," \t\n",&lasts);
       
       i++;
     }

     if (debugLevel)
        printf("reloadUdpTable: local col:%d rem col:%d\n",
               localCol,remCol);

     /* if we couldnt find both columns, give up */
     if ((localCol == 0) || (remCol == 0) || (stateCol == 0)) {
        printf("reloadUdpTable: unable to discern cols %d, %d\n",
                localCol,remCol);
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pre-allocate some space if we need to */
     enlarge_udp_table();

     /* now, loop through rest of file pullingout udp socket details */
     numUdps = 0;

     while (fgets(line,sizeof(line)-1,fp) != NULL) {

       udpTable[numUdps].ipVersion = 4;
       udpTable[numUdps].localIPv6[0] = 0;
       udpTable[numUdps].remIPv6[0] = 0;

       i=0;
       aPtr = line;
       aPtr = strtok_r(aPtr," \t\n",&lasts);
       while (aPtr != NULL) {
         if (i == localCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%08x:%04x",
                  &udpTable[numUdps].localIp,
                  &udpTable[numUdps].localPort);
         }
         if (i == remCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%08x:%04x",
                  &udpTable[numUdps].remIp,
                  &udpTable[numUdps].remPort);
         }
         if (i == stateCol) {
 	    strcpy(aToken,aPtr);
            sscanf(aToken,"%x",&sockState);
            udpTable[numUdps].state = (u_char)sockState;

            udpTable[numUdps].state = 
              convertUDPState(udpTable[numUdps].state);

         }
         if (i == inodeCol) {
	    strcpy(aToken,aPtr);
            sscanf(aToken,"%llu",&udpTable[numUdps].inode);
         }

         aPtr = strtok_r(NULL," \t\n",&lasts);
         i++;

       } /* while strtok through the line */

       xmpConvertIpAddr(udpTable[numUdps].localIp,ipstr1,sizeof(ipstr1));
       xmpConvertIpAddr(udpTable[numUdps].remIp,ipstr2,sizeof(ipstr2));

       if (debugLevel)
	 printf("reloadUdpTable: State:%02x %s:%d %s:%d, inode:%llu\n",
                 sockState,ipstr1,udpTable[numUdps].localPort,
                 ipstr2,udpTable[numUdps].remPort,
                 udpTable[numUdps].inode);

       numUdps++;
       enlarge_udp_table(); /* enlarge if we need more space */

     } /* fgets line */

     fclose(fp);

     /* ********************************************************* */
     /* now, open/parse /proc/net/udp6 since many linux/linux
        connections are now defaulting over IPv6 */

     if ((fp = fopen("/proc/net/udp6","r")) == NULL) {
       ;
     }
     else {

        /* read first line; it should be column headers; find out
           which columns are local and remote addresses */
        localCol = remCol = stateCol = inodeCol = 0;
        if (fgets(line,sizeof(line)-1,fp) == NULL) {
	   perror("reloadUdpTable: empty udp6 file");
        }
        else {
           i = 0;
           aPtr = line;
           aPtr = strtok_r(aPtr," \t\n",&lasts);
           while (aPtr != NULL) {
             if (strcasecmp(aPtr,"local_address") == 0) {
                localCol = i;
             }
             if (strcasecmp(aPtr,"remote_address") == 0) {
                remCol = i;
             }
             if (strcasecmp(aPtr,"st") == 0) {
         	stateCol = i;
             }
             if (strcasecmp(aPtr,"inode") == 0) {
	        inodeCol = i-2; /* magic -2 */
             }
             aPtr = strtok_r(NULL," \t\n",&lasts);
             i++;
           } /* aPtr */

        } /* fgets first line */

        while (fgets(line,sizeof(line)-1,fp) != NULL) {

          udpTable[numUdps].localIp = 0;
          udpTable[numUdps].remIp = 0;
          udpTable[numUdps].ipVersion = 6;

          i=0;
          aPtr = line;
          aPtr = strtok_r(aPtr," \t\n",&lasts);
          while (aPtr != NULL) {
            if (i == localCol) {
      	       strcpy(aToken,aPtr);
               sscanf(aToken,"%32s:%04x",
                      ipv6addr,
                      &udpTable[numUdps].localPort);
               convertIPv6ByteStr(ipv6addr,
                                  udpTable[numUdps].localIPv6,
                                  48);
            }
            if (i == remCol) {
               strcpy(aToken,aPtr);
               sscanf(aToken,"%32s:%04x",
                      ipv6addr,
                      &udpTable[numUdps].remPort);
               convertIPv6ByteStr(ipv6addr,
                                  udpTable[numUdps].remIPv6,
                                  48);

            }
            if (i == stateCol) {
 	       strcpy(aToken,aPtr);
               sscanf(aToken,"%x",&sockState);
               udpTable[numUdps].state = (u_char)sockState;
               udpTable[numUdps].state = 
                  convertUDPState(udpTable[numUdps].state);
            }
            if (i == inodeCol) {
	       strcpy(aToken,aPtr);
               sscanf(aToken,"%llu",&udpTable[numUdps].inode);
            }

            aPtr = strtok_r(NULL," \t\n",&lasts);
            i++;

          } /* while strtok through the line */

          numUdps++;
          enlarge_udp_table(); /* enlarge if we need more space */

        } /* while udp6 entries */

        fclose(fp);
     }
  
     /* sort? */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadUdpTable: %d udp endpoints found\n",numUdps);

     udpTableTime = time(NULL);
}

/* ************************************************************* */
/* load interface info from /proc/net/dev
   Interface file is not as well-formed as udp/tcp files
   The first two lines are column headers with the first column having
   interface name:bytes joined together rather than by space
   Consequently, this file has to be scrutinized with each linux kernel
   release
*/

static void reloadIfTable()
{
     FILE *fp;
     char line[4096];
     char *aPtr, aStr[256];
     int i,s,s6;
     struct ifreq ifr;
     struct sockaddr *sa;
     struct sockaddr_in *sai;
     struct sockaddr_in6 *sai6;
     int numLoopbacks, numNetIfs;
     static int firstTime = 1;
     char ipv6addr[64];
     char a1[6],a2[6],a3[6],a4[6],a5[6],a6[6],a7[6],a8[7];
     int status;

     /* get lock on ifTable */
     pthread_mutex_lock(&mib2_lock);

     if ((fp = fopen("/proc/net/dev","r")) == NULL) {
        perror("reloadIfTable: fopen dev failed");
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* eat first line */
     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadIfTable: empty dev file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* eat second line */
     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadIfTable: empty dev file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pre-allocate some space if we need to */
     enlarge_if_table();

     /* now, loop through rest of file pulling out interface details */
     numIfs = 0;
     numLoopbacks = numNetIfs = 0;

     while (fgets(line,sizeof(line)-1,fp) != NULL) {

       /* take the colon and put a space there */
       /* XXX for virtual interfaces, we need to re-examine 
          whether we take the first colon or the last colon */
       aPtr = strchr(line,':');
       if (aPtr != NULL) { *aPtr = ' '; }

       /* XXX very very fragile, brittle code */
       sscanf(line,"%s %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld",
              ifTable[numIfs].ifname,
              &ifTable[numIfs].inbytes,
              &ifTable[numIfs].inpackets,
              &ifTable[numIfs].inerrors,
              &ifTable[numIfs].indrops,
              &ifTable[numIfs].infifo,
              &ifTable[numIfs].inframe,
              &ifTable[numIfs].incompressed,
              &ifTable[numIfs].inmulticast,
              &ifTable[numIfs].outbytes,
              &ifTable[numIfs].outpackets,
              &ifTable[numIfs].outerrors,
              &ifTable[numIfs].outdrops,
              &ifTable[numIfs].outfifo,
              &ifTable[numIfs].outcollisions,
              &ifTable[numIfs].outcarrier,
              &ifTable[numIfs].outcompressed);

       /* figure out aliases for easier polling */
       ifTable[numIfs].alias[0] = 0; /* set default */

       if (strcmp(ifTable[numIfs].ifname,"lo") != 0) {
	  sprintf(ifTable[numIfs].alias,"if%d",numNetIfs);
          numNetIfs++;
       }

       DBG2("reloadIfTable: %s aliased to %s\n",
	     ifTable[numIfs].ifname,
	     ifTable[numIfs].alias);

       /* put into appropriate "std" variables -- need to find 
          extra docs on exact definition of some of these
          OS values  XXX
        */
       ifTable[numIfs].ifInOctets = ifTable[numIfs].inbytes;
       ifTable[numIfs].ifInUcastPkts = ifTable[numIfs].inpackets;
       ifTable[numIfs].ifInNUcastPkts = ifTable[numIfs].inmulticast;
       ifTable[numIfs].ifInDiscards = ifTable[numIfs].indrops +
	                              ifTable[numIfs].infifo;

       ifTable[numIfs].ifInErrors = ifTable[numIfs].inerrors +
	                            ifTable[numIfs].inframe;

       ifTable[numIfs].ifInUnknownProtos = 0; /* XXX*/

       ifTable[numIfs].ifOutOctets = ifTable[numIfs].outbytes;
       ifTable[numIfs].ifOutUcastPkts = ifTable[numIfs].outpackets;

       ifTable[numIfs].ifOutNUcastPkts = 0; /* XXX*/

       ifTable[numIfs].ifOutDiscards = ifTable[numIfs].outdrops +
                                       ifTable[numIfs].outfifo;

       ifTable[numIfs].ifOutErrors = ifTable[numIfs].outerrors +
                                     ifTable[numIfs].outcarrier +
                                     ifTable[numIfs].outcollisions;


       /* dont forget to set ifIndex even though we dont use
          it for XMP other than link Up/Down Traps */
       ifTable[numIfs].ifIndex = numIfs+1;

       if (debugLevel)
          printf("reloadIfTable: %s %lld %lld %lld %lld\n",
                ifTable[numIfs].ifname,
                ifTable[numIfs].inbytes,
                ifTable[numIfs].inpackets,
                ifTable[numIfs].outbytes,
                ifTable[numIfs].outpackets);

       numIfs++;
       enlarge_if_table();
     }
     fclose(fp);
 
     if (debugLevel)
        printf("reloadifTable: %d interfaces found\n",numIfs);

     /* now, we should use Linux specific socket code to go
        and get flags, MTUs, etc. for each interface */
     /* SIOCGIFNAME, SIOCGIFCONF, SIOCGIFFLAGS, SIOCGIFADDR */

     if ((s = socket(PF_INET,SOCK_DGRAM,0)) < 0) {
        perror("reloadIfTable: socket call failed");
     }

     i = 0;
     while (i < numIfs) {

       /* for loopback interface, we should not pass lo0 as name
          but simply lo */

       /* get flags */
       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);

       DBG1("Examining interface '%s'\n",ifTable[i].ifname);

       if (ioctl(s,SIOCGIFFLAGS,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFFLAGS failed for %s %d\n",
                ifr.ifr_name,errno);
       }
       else {
          ifTable[i].flags = (u_int)ifr.ifr_flags;
       }

       /* verify OperStatus and AdminStatus; we may need
          to switch flag they look at; also, look at oper status
          to see if we should send a linkUp or linkDown trap */

       /* we're assuming that the interfaces come out of the API
          (or file) in the same order as last time XXX */
          
       ifTable[i].oldIfOperStatus = ifTable[i].ifOperStatus;

       ifTable[i].ifOperStatus = 2;
       if (ifTable[i].flags & IFF_RUNNING)
          ifTable[i].ifOperStatus = 1;

       ifTable[i].ifAdminStatus = 2;
       if (ifTable[i].flags & IFF_UP)
          ifTable[i].ifAdminStatus = 1; 

       /* get MTU */
       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       ifr.ifr_mtu = 0;
       if (ioctl(s,SIOCGIFMTU,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFMTU failed for %s, %d\n",
                ifr.ifr_name,errno);
       }
       else {
          ifTable[i].mtu = (u_int)ifr.ifr_mtu;
       }
 
       /* get Tx QLen */
       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       if (ioctl(s,SIOCGIFTXQLEN,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFTXQLEN failed for %s, %d\n",
                ifr.ifr_name,errno);
       }
       else {
          ifTable[i].ifOutQLen = (u_int)ifr.ifr_qlen;
       }

       /* bandwidth -- ifr_bandwidth */
       ifTable[i].ifSpeed = 1; /* XXX prevent divide by zero in managers */
       /* ftp://ftp.scyld.com/pub/diag/mii-diag.c */

       /* get HW addr */
       ifTable[i].hwaddrlen = 0;
       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       if (ioctl(s,SIOCGIFHWADDR,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFHWADDR failed for %s, %d\n",
                ifr.ifr_name,errno);
       }
       else {
          ifTable[i].hwaddrlen = 6; /* XXX */
          bcopy(ifr.ifr_hwaddr.sa_data,ifTable[i].hwaddr,6);
       }

       /* get ifType XXX */
       ifTable[i].ifType = 1; /* other(1) */

       if (strncmp(ifTable[i].ifname,"lo",2) == 0) {
 	  ifTable[i].ifType = 24; /* softwareloopback(24) */
          /* if its loopback addr, set hwaddrlen to 0 */
          ifTable[i].hwaddrlen = 0;
       }
       if (strncmp(ifTable[i].ifname,"eth",3) == 0) {
 	  ifTable[i].ifType = 6; /* (6) */
       }
       if (strncmp(ifTable[i].ifname,"em",3) == 0) {
 	  ifTable[i].ifType = 6; /* (6) */
       }
       if (strncmp(ifTable[i].ifname,"tr",2) == 0) {
 	  ifTable[i].ifType = 9;  /* tokenring(9) */
       }
       if (strncmp(ifTable[i].ifname,"ppp",3) == 0) {
 	  ifTable[i].ifType = 23;  /* ppp(23) */

          /* ppp interfaces also do not have a mac layer addr
             since they tunnel over some physical interface */
	  ifTable[i].hwaddrlen = 0;
       }
       if (strncmp(ifTable[i].ifname,"sl",2) == 0) {
 	  ifTable[i].ifType = 28;  /* slip(28) */
       }
       if (strncmp(ifTable[i].ifname,"sit",3) == 0) {
	  ifTable[i].ifType = 215;  /* 6to4 IPv6-in-IPv4 */
       }

       /* we can get interfaces ip-addr, broadcast addr, netmask if we want */

       /* defaults */
       ifTable[i].ipv6Addr[0] = 0;
       ifTable[i].prefixLen = 0;
       ifTable[i].ipv4addr = 0;
       ifTable[i].ipv4mask = 0;
       ifTable[i].ipv4bcast = 0;
       ifTable[i].ipVersion = 4; /* default */

       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       if (ioctl(s,SIOCGIFADDR,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFADDR failed for '%s',%d\n",
                ifr.ifr_name,errno);
       }
       else {
 	  sa = (struct sockaddr *)&ifr.ifr_addr;          
 	  sai = (struct sockaddr_in *)&ifr.ifr_addr;


          if (sa->sa_family == AF_INET) {
	     ifTable[i].ipVersion = 4;
             ifTable[i].ipv4addr = (u_int)sai->sin_addr.s_addr;
             xmpConvertIpAddr(ifTable[i].ipv4addr,aStr,255);
             if (debugLevel)
                printf("%s addr %s\n",ifTable[i].ifname,aStr);
          }
          else {
	     printf("Found %d family address for '%s'\n",
                    sa->sa_family,ifTable[i].ifname);
          }
       }

       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       if (ioctl(s,SIOCGIFBRDADDR,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFBRDADDR failed for '%s',%d\n",
               ifr.ifr_name,errno);
       }
       else {
 	  sa = (struct sockaddr *)&ifr.ifr_broadaddr;
 	  sai = (struct sockaddr_in *)&ifr.ifr_broadaddr;
          if (sa->sa_family == AF_INET) {
             ifTable[i].ipv4bcast = (u_int)sai->sin_addr.s_addr;
             xmpConvertIpAddr(ifTable[i].ipv4bcast,aStr,255);
             if (debugLevel)
                printf("%s bcast addr %s\n",ifTable[i].ifname,aStr);
	  }
       }

       ifTable[i].ipv4mask = 0;
       sprintf(ifr.ifr_name,"%s",ifTable[i].ifname);
       if (ioctl(s,SIOCGIFNETMASK,&ifr) < 0) {
	  DBG2("interface ioctl SIOCGIFNETMASK failed for '%s',%d\n",
                  ifr.ifr_name,errno);
       }
       else {
 	  sa = (struct sockaddr *)&ifr.ifr_netmask;
 	  sai = (struct sockaddr_in *)&ifr.ifr_netmask;
          if (sa->sa_family == AF_INET) {
             ifTable[i].ipv4mask = (u_int)sai->sin_addr.s_addr;
             xmpConvertIpAddr(ifTable[i].ipv4mask,aStr,255);
             if (debugLevel)
                printf("%s mask %s\n",ifTable[i].ifname,aStr);
	  }
       }

       /* if this is not our first time, compare old and new ifOperStatus */
       if (firstTime != 1) {

	  if (ifTable[i].oldIfOperStatus != ifTable[i].ifOperStatus) {

	     if (ifTable[i].ifOperStatus == 2) {
	        sendLinkDownTrap(ifTable[i].ifIndex,
			         ifTable[i].ifAdminStatus,
			         ifTable[i].ifOperStatus,
                                 ifTable[i].ifname);
	     }
             if (ifTable[i].ifOperStatus == 1) {
	        sendLinkUpTrap(ifTable[i].ifIndex,
			       ifTable[i].ifAdminStatus,
			       ifTable[i].ifOperStatus,
                               ifTable[i].ifname);
             }

	  } /* if old ! = new */

       } /* if !firstTime */

       /* now that we have obtained info for the interface,
          if its a loopback interface, add an instance onto the name */
       if (strstr(ifTable[i].ifname,"lo") == ifTable[i].ifname) {
	  sprintf(ifTable[i].ifname,"lo%d",numLoopbacks);
          numLoopbacks++;
       }
      
       i++;
     }
     close(s);

     /* go look through /proc/net/if_inet6 for ipv6 interfaces */

     /* format is: addr(%4s...), index/%02x, prefixlen/%02x, scope/%02x, 
                   status/%02x, devname %20s */
     /* scope=0,global, linklocal, sitelocal, etc */

     if ((fp = fopen("/proc/net/if_inet6","r")) != NULL) {
 
        while (fgets(line,sizeof(line)-1,fp) != NULL) {

          enlarge_if_table();

          bzero((void *)&ifTable[numIfs],sizeof(mib2IfEntry_t));

          ifTable[numIfs].ipVersion = 6;
          ifTable[numIfs].ifType = 1; /* other(1) */
          ifTable[numIfs].ifSpeed = 1; /* default XXX */

          sscanf(line,"%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %s",
                 a1,a2,a3,a4,a5,a6,a7,a8,
                 &ifTable[numIfs].ifIndex,
                 &ifTable[numIfs].prefixLen,
                 &ifTable[numIfs].scope,
                 &status,
                 aStr);
          sprintf(ifTable[numIfs].ifname,"%s(v6)",aStr);

          /* ipv6 addrs come out as the hex digits w/o any bit twiddling */
          sprintf(ifTable[numIfs].ipv6Addr,
                  "%04s:%04s:%04s:%04s:%04s:%04s:%04s:%04s",
                  a1,a2,a3,a4,a5,a6,a7,a8);

          DBG1("Found ipv6 addr %s\n",ifTable[numIfs].ipv6Addr);

	  numIfs++;
        }
        fclose(fp);

     } /* if fopen /proc/net/if_inet6 */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadifTable: finished, %d\n",numIfs);

     ifTableTime = time(NULL);
     firstTime = 0;
}

/* ************************************************************* */
/* load arp table from /proc/net/arp; IP addresses are in ascii dotted-quad
   notation while type, flags, etc are in hex; the hardware addr is
   in typical format with colon-separated hex byte values */

static void reloadArpTable()
{
     FILE *fp;
     char line[4096], aStr1[64], aStr2[64], aStr3[64];

     /* get lock on routeTable */
     pthread_mutex_lock(&mib2_lock);

     if ((fp = fopen("/proc/net/arp","r")) == NULL) {
        perror("reloadArpTable: fopen arp failed");
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pull out column headers XXX for portability, we should do this. 
        for now, we assume the headers wont change */

     /* for now, just eat first line */
     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadArpTable: empty arp file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pre-allocate some space if we need to */
     enlarge_arp_table();
     numArps = 0;

     while (fgets(line,sizeof(line)-1,fp) != NULL) {

       /* XXX very very fragile, brittle code */
       sscanf(line,"%s %x %x %s %s %s",
              aStr1,  /* ipaddr */
              &arpTable[numArps].hwtype,
              &arpTable[numArps].flags,
              aStr2,  /* hwaddr */
              aStr3,  /* mask */
              arpTable[numArps].ifname);

       arpTable[numArps].ipv4addr = (u_int)inet_addr(aStr1);
       
       /* make assumption that arp hwaddr is 6 bytes long XXX */
       arpTable[numArps].hwaddrlen = 6;
       sscanf(aStr2,"%x:%x:%x:%x:%x:%x",
              &arpTable[numArps].hwaddr[0],
              &arpTable[numArps].hwaddr[1],
              &arpTable[numArps].hwaddr[2],
              &arpTable[numArps].hwaddr[3],
              &arpTable[numArps].hwaddr[4],
              &arpTable[numArps].hwaddr[5]);

       /* dont worry about arp mask for now XXX */

       if (debugLevel)
          printf("Arp: %s/%lx -> %02x:%02x:%02x:%02x:%02x:%02x if:%s\n",
	         aStr1,
                 arpTable[numArps].ipv4addr,
                 arpTable[numArps].hwaddr[0],
                 arpTable[numArps].hwaddr[1],
                 arpTable[numArps].hwaddr[2],
                 arpTable[numArps].hwaddr[3],
                 arpTable[numArps].hwaddr[4],
                 arpTable[numArps].hwaddr[5],
                 arpTable[numArps].ifname);

       numArps++;
       enlarge_arp_table();
     }
     fclose(fp);


     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadArpTable: finished, %d\n",numArps);

     arpTableTime = time(NULL);
}

/* ************************************************************* */
/* load route table from /proc/net/route */
/* dest, gateway, mask appear to be in hex host byte order;
   however, when we read them in, they get byte-swapped into network order
   so we're fine to deal with them w/o swapping */

static void reloadRouteTable()
{
     FILE *fp;
     char line[4096];
     char aStr1[256], aStr2[256], aStr3[256];

     /* get lock on routeTable */
     pthread_mutex_lock(&mib2_lock);

     if ((fp = fopen("/proc/net/route","r")) == NULL) {
        perror("reloadRouteTable: fopen route failed");
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pull out column headers XXX for portability, we should do this. 
        for now, we assume the headers wont change */

     /* for now, just eat first line */
     if (fgets(line,sizeof(line)-1,fp) == NULL) {
        perror("reloadRouteTable: empty route file");
        fclose(fp);
        /* release lock */
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* pre-allocate some space if we need to */
     enlarge_route_table();

     /* now, loop through rest of file pulling out route details */
     numRoutes = 0;
     while (fgets(line,sizeof(line)-1,fp) != NULL) {

       /* XXX very very fragile, brittle code */
       sscanf(line,"%s %08x %08x %x %d %d %d %08x",
              routeTable[numRoutes].ifname,
              &routeTable[numRoutes].ipv4dest,
              &routeTable[numRoutes].ipv4gway,
              &routeTable[numRoutes].flags,
              &routeTable[numRoutes].refcnt,
              &routeTable[numRoutes].use,
              &routeTable[numRoutes].routeMetric,
              &routeTable[numRoutes].ipv4mask);

       /* set routeType */
       routeTable[numRoutes].routeType = 1; /* other */
       
       /* route proto */
       routeTable[numRoutes].routeProto = 1; /* other */
       if (routeTable[numRoutes].flags & RTF_STATIC)
          routeTable[numRoutes].routeProto = 1; /* other/static */
       if (routeTable[numRoutes].flags & RTF_DYNAMIC)
          routeTable[numRoutes].routeProto = 4; /* other */


       if (debugLevel) {
          xmpConvertIpAddr(routeTable[numRoutes].ipv4dest,aStr1,255);
          xmpConvertIpAddr(routeTable[numRoutes].ipv4gway,aStr2,255);
          xmpConvertIpAddr(routeTable[numRoutes].ipv4mask,aStr3,255);

          printf("Route: if:%s dst:%s gway:%s mask:%s metric:%d\n",
                 routeTable[numRoutes].ifname,
	         aStr1,aStr2,aStr3,routeTable[numRoutes].routeMetric);
       } /* debugLevel */

       numRoutes++;
       enlarge_route_table();
     }
     fclose(fp);

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadRouteTable: finished, %d\n",numRoutes);

     routeTableTime = time(NULL);
}

/* ************************************************************* */
/* reload IP Addr table -- basically, we use the ifTable since
   we've pre-fetched the info needed and put it in ifTable 

   This plan doesnt work due to virtual interfaces and the fact
   that one interface could have multiple IP addresses XXX
*/
 

static void reloadIpAddrTable()
{
     /* copy the contents of ifTable for now */

     /* get lock on routeTable */
     pthread_mutex_lock(&mib2_lock);

     /* pre-allocate some space if we need to */
     enlarge_addr_table();

     numAddrs = 0;

     while (numAddrs < numIfs) {

       strcpy(addrTable[numAddrs].ifname,ifTable[numAddrs].ifname);
       addrTable[numAddrs].ipv4ReasmMaxSize = 1; /* XXX */

       if (ifTable[numAddrs].ipVersion == 4) {

	  addrTable[numAddrs].ipVersion = 4;
          addrTable[numAddrs].ipv4addr = ifTable[numAddrs].ipv4addr;
          addrTable[numAddrs].ipv4mask = ifTable[numAddrs].ipv4mask;
          addrTable[numAddrs].ipv6Addr[0] = 0;

          if (ifTable[numAddrs].ipv4bcast & 0x01)
       	     addrTable[numAddrs].ipv4bcast = 1;
          else
 	     addrTable[numAddrs].ipv4bcast = 0;
       }
       else {
	  addrTable[numAddrs].ipv4addr = 0;
          addrTable[numAddrs].ipv4mask = 0;
          addrTable[numAddrs].ipVersion = 6;
          addrTable[numAddrs].ipv4bcast = 0;
 	  strcpy(addrTable[numAddrs].ipv6Addr,
                 ifTable[numAddrs].ipv6Addr);
          addrTable[numAddrs].ipv6PrefixLen = ifTable[numAddrs].prefixLen;
          addrTable[numAddrs].ipv6AddrType = ifTable[numAddrs].scope;
       }

       numAddrs++;
       enlarge_addr_table();
     }

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     return;
}

/* ************************************************************* */

static void ipStatString(char *header, char *valstr, ipStats_t *i)
{
     if ((header == NULL) || (valstr == NULL) || (i == NULL))
        return;

     if (strcasecmp(header,"Forwarding") == 0)
        i->ipForwarding = (int)atoi(valstr);
     if (strcasecmp(header,"DefaultTTL") == 0)
        i->ipDefaultTTL = (int)atoi(valstr);

     if (strcasecmp(header,"InReceives") == 0)
        i->ipInReceives = (u_int64_t)strtoull(valstr,NULL,10);

     if (strcasecmp(header,"InHdrErrors") == 0)
        i->ipInHdrErrors = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InAddrErrors") == 0)
        i->ipInAddrErrors = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"ForwDatagrams") == 0)
        i->ipForwDatagrams = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InUnknownProtos") == 0)
        i->ipInUnknownProtos = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InDiscards") == 0)
        i->ipInDiscards = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InDelivers") == 0)
        i->ipInDelivers = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutRequests") == 0)
        i->ipOutRequests = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutDiscards") == 0)
        i->ipOutDiscards = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutNoRoutes") == 0)
        i->ipOutNoRoutes = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"ReasmTimeout") == 0)
        i->ipReasmTimeout = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"ReasmReqds") == 0)
        i->ipReasmReqds = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"ReasmOKs") == 0)
        i->ipReasmOKs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"ReasmFails") == 0)
        i->ipReasmFails = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"FragOKs") == 0)
        i->ipFragOKs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"FragFails") == 0)
        i->ipFragFails = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"FragCreates") == 0)
        i->ipFragCreates = (u_int64_t)strtoull(valstr,NULL,10);

     /* does not look like Linux supports this one XXX */
     if (strcasecmp(header,"RoutingDiscards") == 0)
        i->ipRoutingDiscards = (u_int64_t)strtoull(valstr,NULL,10);

     return;
}

static void icmpStatString(char *header, char *valstr, icmpStats_t *i)
{
     if ((header == NULL) || (valstr == NULL) || (i == NULL))
        return;

     if (strcasecmp(header,"InMsgs") == 0)
        i->icmpInMsgs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InErrors") == 0)
        i->icmpInErrors = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InDestUnreachs") == 0)
        i->icmpInDestUnreachs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InTimeExcds") == 0)
        i->icmpInTimeExcds = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InParmProbs") == 0)
        i->icmpInParmProbs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InSrcQuenchs") == 0)
        i->icmpInSrcQuenchs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InRedirects") == 0)
        i->icmpInRedirects = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InEchos") == 0)
        i->icmpInEchos = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InEchoReps") == 0)
        i->icmpInEchoReps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InTimestamps") == 0)
        i->icmpInTimestamps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InTimestampReps") == 0)
        i->icmpInTimestampReps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InAddrMasks") == 0)
        i->icmpInAddrMasks = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InAddrMaskReps") == 0)
        i->icmpInAddrMaskReps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutMsgs") == 0)
        i->icmpOutMsgs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutErrors") == 0)
        i->icmpOutErrors = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutDestUnreachs") == 0)
        i->icmpOutDestUnreachs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutTimeExcds") == 0)
        i->icmpOutTimeExcds = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutParmProbs") == 0)
        i->icmpOutParmProbs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutSrcQuenchs") == 0)
        i->icmpOutSrcQuenchs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutRedirects") == 0)
        i->icmpOutRedirects = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutEchos") == 0)
        i->icmpOutEchos = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutEchoReps") == 0)
        i->icmpOutEchoReps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutTimestamps") == 0)
        i->icmpOutTimestamps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutTimestampReps") == 0)
        i->icmpOutTimestampReps = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutAddrMasks") == 0)
        i->icmpOutAddrMasks = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutAddrMaskReps") == 0)
        i->icmpOutAddrMaskReps = (u_int64_t)strtoull(valstr,NULL,10);

     return;
}

static void tcpStatString(char *header, char *valstr, tcpStats_t *t)
{
     if ((header == NULL) || (valstr == NULL) || (t == NULL))
        return;

     if (strcasecmp(header,"RtoAlgorithm") == 0)
        t->tcpRtoAlgorithm = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"RtoMin") == 0)
        t->tcpRtoMin = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"RtoMax") == 0)
        t->tcpRtoMax = (u_int64_t)strtoull(valstr,NULL,10);

     if (strcasecmp(header,"MaxConn") == 0)
        t->tcpMaxConn = atoi(valstr);

     if (strcasecmp(header,"ActiveOpens") == 0)
        t->tcpActiveOpens = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"PassiveOpens") == 0)
        t->tcpPassiveOpens = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"AttemptFails") == 0)
        t->tcpAttemptFails = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"EstabResets") == 0)
        t->tcpEstabResets = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"CurrEstab") == 0)
        t->tcpCurrEstab = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InSegs") == 0)
        t->tcpInSegs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutSegs") == 0)
        t->tcpOutSegs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"RetransSegs") == 0)
        t->tcpRetransSegs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InErrs") == 0)
        t->tcpInErrs = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutRsts") == 0)
        t->tcpOutRsts = (u_int64_t)strtoull(valstr,NULL,10);

     return;
}

static void udpStatString(char *header, char *valstr, udpStats_t *u)
{
     if ((header == NULL) || (valstr == NULL) || (u == NULL))
        return;

     if (strcasecmp(header,"InDatagrams") == 0)
       u->udpInDatagrams = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"NoPorts") == 0)
        u->udpNoPorts = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"InErrors") == 0)
        u->udpInErrors = (u_int64_t)strtoull(valstr,NULL,10);
     if (strcasecmp(header,"OutDatagrams") == 0)
        u->udpOutDatagrams = (u_int64_t)strtoull(valstr,NULL,10);

     return;
}

static void ipv6StatString(char *header, char *valstr)
{

#if 0

Ip6InReceives                   	3842382
Ip6InHdrErrors                  	0
Ip6InTooBigErrors               	0
Ip6InNoRoutes                   	0
Ip6InAddrErrors                 	0
Ip6InUnknownProtos              	0
Ip6InTruncatedPkts              	0
Ip6InDiscards                   	0
Ip6InDelivers                   	2906499
Ip6OutForwDatagrams             	0
Ip6OutRequests                  	2907123
Ip6OutDiscards                  	0
Ip6OutNoRoutes                  	1909
Ip6ReasmTimeout                 	0
Ip6ReasmReqds                   	0
Ip6ReasmOKs                     	0
Ip6ReasmFails                   	0
Ip6FragOKs                      	0
Ip6FragFails                    	0
Ip6FragCreates                  	0
Ip6InMcastPkts                  	944487
Ip6OutMcastPkts                 	289
Ip6InOctets                     	2807765406
Ip6OutOctets                    	2567632237
Ip6InMcastOctets                	197267130
Ip6OutMcastOctets               	20976
Ip6InBcastOctets                	0
Ip6OutBcastOctets               	0
Icmp6InMsgs                     	9644
Icmp6InErrors                   	11
Icmp6OutMsgs                    	1611
Icmp6OutErrors                  	0
Icmp6InDestUnreachs             	220
Icmp6InPktTooBigs               	14
Icmp6InTimeExcds                	0
Icmp6InParmProblems             	0
Icmp6InEchos                    	60
Icmp6InEchoReplies              	135
Icmp6InGroupMembQueries         	0
Icmp6InGroupMembResponses       	0
Icmp6InGroupMembReductions      	0
Icmp6InRouterSolicits           	0
Icmp6InRouterAdvertisements     	8310
Icmp6InNeighborSolicits         	622
Icmp6InNeighborAdvertisements   	283
Icmp6InRedirects                	0
Icmp6InMLDv2Reports             	0
Icmp6OutDestUnreachs            	196
Icmp6OutPktTooBigs              	0
Icmp6OutTimeExcds               	0
Icmp6OutParmProblems            	0
Icmp6OutEchos                   	265
Icmp6OutEchoReplies             	60
Icmp6OutGroupMembQueries        	0
Icmp6OutGroupMembResponses      	0
Icmp6OutGroupMembReductions     	0
Icmp6OutRouterSolicits          	3
Icmp6OutRouterAdvertisements    	0
Icmp6OutNeighborSolicits        	460
Icmp6OutNeighborAdvertisements  	622
Icmp6OutRedirects               	0
Icmp6OutMLDv2Reports            	5
Icmp6InType1                    	220
Icmp6InType2                    	14
Icmp6InType128                  	60
Icmp6InType129                  	135
Icmp6InType134                  	8310
Icmp6InType135                  	622
Icmp6InType136                  	283
Icmp6OutType1                   	196
Icmp6OutType128                 	265
Icmp6OutType129                 	60
Icmp6OutType133                 	3
Icmp6OutType135                 	460
Icmp6OutType136                 	622
Icmp6OutType143                 	5
Udp6InDatagrams                 	2550504
Udp6NoPorts                     	66
Udp6InErrors                    	0
Udp6OutDatagrams                	2550570
Udp6RcvbufErrors                	0
Udp6SndbufErrors                	0
UdpLite6InDatagrams             	0
UdpLite6NoPorts                 	0
UdpLite6InErrors                	0
UdpLite6OutDatagrams            	0
UdpLite6RcvbufErrors            	0
UdpLite6SndbufErrors            	0

#endif

}

/* ************************************************************* */
/* reload stats for ip, icmp, tcp, udp from /proc/net/snmp
   the file contains a series of lines where first token identifies
   the protocol followed by either column headers or the actual stats
   themselves.  Lines come in pairs: first line is columns, second line is 
   stats.  Stats appear to be in base-10.
*/

static void reloadStats()
{
     FILE *fp;
     char line1[4096], line2[4096];
     char *lasts1, *lasts2;
     char *aPtr1, *aPtr2, *cptr;
     int ipflag=0, icmpflag=0, tcpflag=0, udpflag=0;

     if (debugLevel)
        printf("reloadStats: starting\n");

     /* we dont lock these stats structs since we are updating
        field at a time and there is no real harm if we are writing
        to a field whilst someone is reading from it */

     if ((fp = fopen("/proc/net/snmp","r")) == NULL) {
        perror("reloadStats: fopen snmp failed");
        return;
     }

     while (fgets(line1,sizeof(line1)-1,fp) != NULL) {

       /* read 2nd line; if not there, break out of loop */
       if (fgets(line2,sizeof(line2)-1,fp) == NULL) {
  	  break;
       }

       /* verify that both line1, line2 have same first token */

       aPtr1 = line1;  /* column headers */
       aPtr2 = line2;  /* stats columns */
       aPtr1 = strtok_r(aPtr1," \t\n",&lasts1);
       aPtr2 = strtok_r(aPtr2," \t\n",&lasts2);
       if (strcasecmp(aPtr1,aPtr2) != 0) {
   	  printf("reloadStats: tokens '%s' '%s' dont match\n",aPtr1,aPtr2);
	  continue; /* with reading pair of lines */
       }

       /* reset flags */
       ipflag = icmpflag = tcpflag = udpflag = 0;

       if (strcasecmp(aPtr1,"Ip:") == 0) ipflag = 1;
       if (strcasecmp(aPtr1,"Icmp:") == 0) icmpflag = 1;
       if (strcasecmp(aPtr1,"Tcp:") == 0) tcpflag = 1;
       if (strcasecmp(aPtr1,"Udp:") == 0) udpflag = 1;

       /* loop through columns of line1/line2 */
       /* read next columns before looping */
       aPtr1 = strtok_r(NULL," \t\n",&lasts1);
       aPtr2 = strtok_r(NULL," \t\n",&lasts2);

       while (aPtr1 && aPtr2) {

         /* strip any newlines off the strings */
         if ((cptr = strchr(aPtr1,'\n')) != NULL)
  	    *cptr = 0;
         if ((cptr = strchr(aPtr2,'\n')) != NULL)
  	    *cptr = 0;

         if (ipflag)
 	    ipStatString(aPtr1,aPtr2,&ipstats);
         if (icmpflag)
 	    icmpStatString(aPtr1,aPtr2,&icmpstats);
         if (tcpflag)
 	    tcpStatString(aPtr1,aPtr2,&tcpstats);
         if (udpflag)
 	    udpStatString(aPtr1,aPtr2,&udpstats);

         /* read next columns */
         aPtr1 = strtok_r(NULL," \t\n",&lasts1);
         aPtr2 = strtok_r(NULL," \t\n",&lasts2);

       } /* done with columns */


     } /* while line1/line2 */

     fclose(fp);

     statsTime = time(NULL);

#if 0
     printf("reloadStats: udp: %lld %lld %lld %lld\n",
	    udpstats.InDatagrams,
	    udpstats.NoPorts, 
	    udpstats.InErrors,
	    udpstats.OutDatagrams);

     printf("reloadStats: icmp: %lld %lld %lld\n",
            icmpstats.InMsgs,icmpstats.OutMsgs,icmpstats.OutAddrMaskReps);

     printf("reloadStats: tcp: %d %lld %lld %lld\n",
	    tcpstats.MaxConn,tcpstats.InSegs,tcpstats.OutSegs,
            tcpstats.OutRsts);

     printf("reloadStats: ip: %d %d %lld %lld %lld\n",
 	     ipstats.ipForwarding,
 	     ipstats.ipDefaultTTL,
 	     ipstats.ipInReceives,
 	     ipstats.ipOutRequests,
 	     ipstats.ipFragCreates);
#endif

     if (debugLevel)
        printf("reloadStats: finished\n");
}

/* ************************************************************* */
/* periodically reload caches */

static void *linuxMib2_dowork(void *arg)
{
     time_t current;
     int i;

     while (mib2ThreadKill != 1) {

       current = time(NULL);

       /* tcp table */
       if ((current - tcpTableTime) > MIB2_CACHE_LIFE) {
	  reloadTcpTable();
       }

       /* udp table */
       if ((current - udpTableTime) > MIB2_CACHE_LIFE) {
	  reloadUdpTable();
       }

       /* if table */
       if ((current - ifTableTime) > MIB2_CACHE_LIFE) {
	  reloadIfTable();
          reloadIpAddrTable();
       }

       /* route table */
       if ((current - routeTableTime) > MIB2_CACHE_LIFE) {
	  reloadRouteTable();
       }

       /* arp table */
       if ((current - arpTableTime) > MIB2_CACHE_LIFE) {
	  reloadArpTable();
       }

       /* ipAddr table -- is pre-fetched in ifTable */

       /* stats: ip, tcp, udp, arp, etc. */
       if ((current - statsTime) > MIB2_CACHE_LIFE) {
	  reloadStats();
       }

       for (i=0; i<2 && mib2ThreadKill == 0; i++) {
           sleep(15);
       }

       DBG("linuxMib2_dowork: working . . . .\n");

     }
 
     printf("linuxMib2: worker thread returning\n");

     return NULL;  /* effectively killing me */
}

/* ************************************************************* */

int start()
{
     int mib2TrapHandler();

     printf("linuxMib2: start called\n");

     /* start a thread */
     if (pthread_create(&wt,NULL,linuxMib2_dowork,NULL) < 0) {
        perror("Failed to create worker thread for linuxMib2");
        return -1;
     }
     pthread_detach(wt);

     xmpRegisterTrapHandler("mib2",mib2TrapHandler);

     return 1;
}

/* ************************************************************* */

#endif /* LINUX */
