/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   THIS SOFTWARE IS THE CONFIDENTIAL AND PROPRIETARY PRODUCT OF 
   KRUPCZAK.ORG, LLC.  ANY UNAUTHORIZED USE, REPRODUCTION, OR
   TRANSFER OF THIS SOFTWARE IS STRICTLY PROHIBITED.  (SUBJECT TO LIMITED
   DISTRIBUTION AND RESTRICTED DISCLOSURE ONLY.)  ALL RIGHTS RESERVED.
*/

/* file - linkedlist.h */
/* rdk@krupczak.org */

/* structures for generic doubly linked list package */

/* ************************************************************* */

typedef struct linkedElement {

     void *data;

     struct linkedElement *next;
     struct linkedElement *prev;

} linkedElement_t;

typedef struct linkedList {
   
     char description[64];

     int autoLock;
     
     linkedElement_t *head;
     linkedElement_t *tail;

     pthread_mutex_t lock;

     /* size data, in bytes, that user will glue into elements */
     int dataSize;

     /* compare function used for insertInOrder; returns -1,0,1 */
     int (*elementCompare)();
     int (*elementPrint)();

     /* for freeing user data; if not defined, we do *not* free
        any user-defined data */
     int (*elementFree)(); 

} linkedList_t;

linkedList_t *newLinkedList(int, void *, void *, void *, char *, int);
int destroyLinkedList(linkedList_t *);
int insertInOrder(linkedList_t *, void *);
int insertHead(linkedList_t *, void *);
int insertTail(linkedList_t *, void *);
int destroyElement(linkedList_t *, linkedElement_t *);
void *popHead(linkedList_t *);
void *popTail(linkedList_t *);
int numberElements(linkedList_t *);
int traverseList(linkedList_t *);
linkedElement_t *findElement(linkedList_t *, void *);
u_char *toArrayCopy(linkedList_t *, int *);
int iterateList(linkedList_t *, int (*)());
int drainLinkedList(linkedList_t *list);
int removeElement(linkedList_t *, linkedElement_t *);
int disconnectElement(linkedList_t *, linkedElement_t *);

#ifdef WIN32
#define bcopy(src,dst,n)  memcpy(dst,src,n)
#define bzero(dst,n)      memset(dst,0,n)
#endif

/* ************************************************************* */
