/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - main.c */
/* rdk@krupczak.org */

/* Daemon's main function and initialization code */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdio.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>

#ifndef WIN32
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include <stdlib.h>

#include "xmp.h"
#include "socklib.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "globals.h"

/* ************************************************************* */
/* process-wide signal handlers; we should have all the
   threads but the main daemon thread block all signals */

#ifndef WIN32

void sigpipe_handler(int sig)
{
     static int firsttime = 0;
     time_t pipeTime;
     char timeBuf[64];

     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
        pipeTime = time(NULL);
        ctime_r(&pipeTime,timeBuf);

        fprintf(stderr,"xmpd: caught SIGPIPE at '%s'\n",timeBuf);

        /* log that we caught the signal */
        if (xmpd_verbose)
 	   fprintf(stderr,"Caught SIGPIPE at '%s'\n",timeBuf);
     }

     /* reschedule myself */
     signal(SIGPIPE,sigpipe_handler);
}

void sigterm_handler(int sig)
{
     static int firsttime = 0;

     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
        /* log that we caught the signal */
        globalShutdown = 1;
     }

     /* reschedule myself */
     signal(SIGTERM,sigterm_handler);
}

void sighup_handler(int sig)
{
     static int firsttime = 0;
     
     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
       /* log that we caught the signal */
     }

     /* reschedule myself */
     signal(SIGHUP,sighup_handler);
     return;
}

void sigusr1_handler(int sig)
{
     static int firsttime = 0;
     
     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
       /* log that we caught the signal */
     }

     /* reschedule myself */
     signal(SIGUSR1,sigusr1_handler);
     return;
}

void sigusr2_handler(int sig)
{
     static int firsttime = 0;
     
     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
       /* log that we caught the signal */
     }

     /* reschedule myself */
     signal(SIGUSR2,sigusr2_handler);
     return;
}

void sigfpe_handler(int sig)
{
     static int firsttime = 0;
     
     if (firsttime == 0) {
        firsttime = 1;
     }
     else {
       /* log that we caught the signal */
     }

     /* reschedule myself */
     signal(SIGFPE,sigfpe_handler);

     /* try this to see if we can avoid a dlopen/pthreads
        spin lock loop 
     signal(SIGFPE,SIG_IGN);
     */

     return;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* make me a UNIX daemon */

#ifndef WIN32

int mini_daemon()
{
     /* handle signal stuff */
     /* SIGUSR1, HUP, FPE, TERM, PIPE? */
     sighup_handler(0);
     sigusr1_handler(0);
     sigusr2_handler(0);
     sigfpe_handler(0);
     sigterm_handler(0);
     sigpipe_handler(0);

     return 0;
}

int make_a_daemon_out_of_me()
{
     pid_t pid;
     struct sigaction sa;

     if ((pid = fork()) < 0) {
        return -1;
     }
     else {
       if (pid != 0)
	  exit(0); /* parent exits */
     }

     /* child continues here */

     setsid();     /* create new session leader */

     chdir("/");   /* change working dir to / to avoid NFS/FS problems */

     chdir(xmpd_home_dir);
     umask(00133); /* clear/set file mode creation mask -rw-r--r-- */

     /* close unnecessary file descriptors so as to not hang shells,
        applications */
     close(STDIN_FILENO);
     close(STDOUT_FILENO);
     close(STDERR_FILENO);

     /* however, we've noticed that on linux, when the rest of the
        code goes to print, we can hang/block the agent since
        fds 0,1,2 may be a socket, a file descriptor, etc.
     */ 
     open("/dev/null",O_RDONLY);
     open("/dev/null",O_WRONLY);
     open("/dev/null",O_WRONLY);

     /* handle signal stuff */
     /* SIGUSR1, HUP, FPE, TERM, PIPE? */
     sighup_handler(0);
     sigusr1_handler(0);
     sigusr2_handler(0);
     sigfpe_handler(0);
     sigterm_handler(0);
     sigpipe_handler(0);

     /* ignore TTY signals for I/O -- SIGTTIN and SIGTTOU */
     signal(SIGTTIN,SIG_IGN);
     signal(SIGTTOU,SIG_IGN);

     /* ignore child processes so as to avoid zombies, defunct
        from any sub-shells we may spawn */
     sa.sa_handler = SIG_IGN;
     sa.sa_flags = SA_NOCLDSTOP;
     sigemptyset(&sa.sa_mask);

     /* on Linux w/o NPTL support, we get the following error
        in syslog:

 smyrna kernel: application bug: xmpd-linux(6026) has SIGCHLD set to SIG_IGN but calls wait().
 smyrna kernel: (see the NOTES section of 'man 2 wait'). Workaround activated.
 smyrna kernel: application bug: xmpd-linux(6026) has SIGCHLD set to SIG_IGN but calls wait().
 smyrna kernel: (see the NOTES section of 'man 2 wait'). Workaround activated.

     */

     sigaction(SIGCHLD,&sa,NULL);

     return 0;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* ************************************************************* */
/* it all starts here on UNIX and WIN32 */

/* on UNIX, set CWD to be home director prior to starting cartographer */
/* or start using absolute path */

int main(int argc, char *argv[])
{
     int i;
     int init_xmpd(), version_info(), license_info();
     int usage(char *);

     /* process args 1..argc-1 only looking for -h 
        or any args that we absolutely need right now */
     for (i=1; i<argc; i++) {
         if (argv[i][0] == '-') {

            switch (argv[i][1]) {

  	      case 's':
 	         win32service = 0; /* run as console application */
                 break;

   	      case 'h':
	         usage(argv[0]);
                 break;

              case 'g':
	 	 dumpcore = 1;
                 break;
          
  	      case 'V':
	 	 version_info();
 	  	 break;

   	      case 'l':
	 	 license_info();
                 break;

            } /* switch */

         } /* if '-' */
     } /* for */

#ifdef WIN32
     if (win32service)
        start_win32_service(argc,argv); /* which then calls init_xmpd() */
     else
        init_xmpd(argc,argv);
#else
     /* UNIX */
     init_xmpd(argc,argv);
#endif

     /* when we return from init_xmpd(), we are done */

     fprintf(stderr,"xmpd: exiting\n");
     exit(0);
}

/* ************************************************************* */
