/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - plugins.h */
/* rdk@krupczak.org */

/* Plugin definitions, APIs, exposed symbols, etc.
   Include this file both within agent and within plugins */

/* ************************************************************* */
/* main definition of a plugin */

typedef struct plugin {

   char modulename[25];            /* well-known name of module */
   char pathname[FILENAME_MAX];    /* where it lives */
   char descr[128];                /* module name/ver/description */

   u_char flags;                   /* always have flags */
#ifndef WIN32
   void *solibhandle;              /* handle for shared-lib */
#else
   HMODULE solibhandle;            /* handle for shared-lib */
#endif

   /* function pointers for invoking module */
#ifndef WIN32
   int    (*init)();               /* initialize library */
   int    (*start)();              /* start/create any threads */
   int    (*finish)();             /* prepare to be unloaded */
   int    (*modinfo)();            /* name, vers/descr */
#else
   FARPROC init;                   /* initialize library */
   FARPROC start;                  /* start/create any threads */
   FARPROC finish;                 /* prepare to be unloaded */
   FARPROC modinfo;                /* name, vers/descr */
#endif

   time_t loadTime;                /* time plugin was loaded */

#if 0   
   struct plugin *next;
#endif

} plugin_t;

/* plugin flags */
#define PLUGIN_EMPTY      0x000
#define PLUGIN_VALID      0x001
#define PLUGIN_UNLOADING  0x002

#define PLUGIN_CLEANUP_TIME   30

int initPlugins();
int findLoadAllPlugins(char *);
int getNumberPlugins();
int unloadAllPlugins();
int cleanupAllPlugins();

char *base64_encode(u_char *, int, int *);
u_char *base64_decode(u_char *, int, int *);

/* ************************************************************* */
/* definitions, symbols, etc. needed by plugins for coding to
   our API spec 
*/

/* symbols in DLLs need to explicitly be exported so
   we define DLLExport macro to declare them for export;
   on UNIX, we do nothing so DLLExport is blank
   Also defined in xmp.h to get around compiler errors
*/

#ifdef WIN32

#define DLLExport   __declspec( dllexport )
#define DLLImport   __declspec ( dllimport )

#else

/* UNIX */
#define DLLExport
#define DLLImport  

#endif /* WIN32 */

/* ************************************************************* */

DLLExport int xmpGetAgentDebugLevel();
DLLExport float xmpGetAgentVersion();
DLLExport int xmpGetAgentPort();
DLLExport int xmpConvertTimeStr(time_t, char *, int);
DLLExport void *xmpGetSymbolValue(char *);
DLLExport int xmpConvertIpAddr(u_int, char *, int);
DLLExport int xmpConvertIpV6Addr(u_char *, char *, int);
DLLExport int xmpAddVar(xmpVar_t *);
DLLExport int xmpRemoveVarByName(char *, char *);
DLLExport u_int xmpHostnameToIpaddr(char *);
DLLExport int xmpHostnameToIpv6addr(char *, void *);
DLLExport int xmpIpVersionOfAddress(char *);
DLLExport int xmpIpVersionOfSocket(int);

DLLExport int xmpdLog(char *);

DLLExport int xmpTransaction(xmpSession_t *, xmpPdu_t *, xmpPdu_t *);
DLLExport int xmpPduCleanup(xmpPdu_t *);
DLLExport int xmpDisconnectSession(xmpSession_t *);
DLLExport int xmpAddressToHostname(char *, char *, int);
DLLExport int xmpIpAddrToHostname(char *, char *, int);
DLLExport int xmpGetSystemName(char *, int);
DLLExport int xmpAddModule(char *, char *, char *, char *, int);
DLLExport int xmpRemoveModule(char *);
DLLExport int xmpCheckPeerModules(char *);
DLLExport float xmpFindBinaryVersion(char *, char *);
DLLExport xmpErrorStatus_t xmpSendTrap(char *, xmpPdu_t *, char *, char *, char *);
DLLExport int xmpGetHomeDir(char *, int);

DLLExport int xmpAddUser(char *, u_int, char *);
DLLExport int xmpSessionSetAuthenUser(xmpSession_t *, char *);

#ifndef WIN32
DLLExport int xmpRegisterTrapHandler(char *, int (*fptr)());
#else
DLLExport int xmpRegisterTrapHandler(char *, FARPROC);
#endif

DLLExport char *xmpBase64Encode(u_char *, int, int *);
DLLExport u_char *xmpBase64Decode(u_char *, int, int *);

/* ************************************************************* */
