/* 
   COPYRIGHT 2007 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - restartxmpd.c */
/* rdk@krupczak.org */

/* restart cartographer daemon and perform any upgrades if necessary */
/* some WinMain code borrowed from trayicon.c by wharfinger */

#include <windows.h>
#include <winsvc.h>

#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>

/* ************************************************************* */

#ifdef XMPDEBUG

#define DBG0(s)
#define DBG(s) printf(s);fflush(stdout)
#define DBG1(s,a) printf(s,a);fflush(stdout)
#define DBG2(s,a,b) printf(s,a,b);fflush(stdout)
#define DBG3(s,a,b,c) printf(s,a,b,c);fflush(stdout)
#define DBG4(s,a,b,c,d) printf(s,a,b,c,d);fflush(stdout)
#define DBG5(s,a,b,c,d,e) printf(s,a,b,c,d,e);fflush(stdout)
#define DBG6(s,a,b,c,d,e,f) printf(s,a,b,c,d,e,f);fflush(stdout)

#else

#define DBG0(s)
#define DBG(s)	
#define DBG1(s,a) 	
#define DBG2(s,a,b)
#define DBG3(s,a,b,c)
#define DBG4(s,a,b,c,d)
#define DBG5(s,a,b,c,d,e)
#define DBG6(s,a,b,c,d,e,f)

#endif

char *versionStr = "KorgVersion 0.56";

#define THIS_CLASSNAME      "Cartographer Restart Window"
#define THIS_TITLE          "Cartographer Restart"

/* my homedir set by winmain() */

char mydir[512];

/* ************************************************************* */

static FILE *fp = NULL;

int logMessage(char *messageStr)
{
     char filename[256];
     char timestr[64];
     char *ptr;
     time_t now;

     if (fp == NULL) {
        sprintf(filename,"restartxmpd.log");
        if ((fp = fopen(filename,"a")) == NULL) {
  	   return -1;
        }
     }

     now = time(NULL);
     ptr = ctime(&now);
     strncpy(timestr,ptr,63);
     timestr[63] = 0;
     if (timestr[strlen(timestr)-1] == '\n')
        timestr[strlen(timestr)-1] = 0;

     if (fp != NULL) {
        fprintf(fp,"%s %s\n",timestr,messageStr);
        fflush(fp);
     }

     return 1;
}

/* ************************************************************* */
/* take .new files and rename the current */

int renameFiles()
{
     HANDLE hFind;
     WIN32_FIND_DATA dent;
     char searchdir[1024];
     char currentName[512],oldName[512], *cptr;
     char message[1024];
     int ret;

     sprintf(searchdir,"%s\\*.new",mydir);

     hFind = FindFirstFile(searchdir,&dent);

     /* look for *.new files */
     while (hFind != INVALID_HANDLE_VALUE) {

       if (strstr(dent.cFileName,".new") == NULL) {
          if (FindNextFile(hFind,&dent) == 0) {
   	     break;
          }
	  continue;
       }

       sprintf(message,"Found file %s",dent.cFileName);
       logMessage(message);
       //MessageBox(NULL,message,THIS_TITLE,MB_ICONERROR|MB_OK|MB_TOPMOST);

       /* construct current name */
       strncpy(currentName,dent.cFileName,511);
       currentName[511] = 0;
       cptr = strstr(currentName,".new");
       if (cptr != NULL) *cptr = 0;

       /* we dont care of there is current; if so, rename it */
       sprintf(oldName,"%s.old",currentName);

       sprintf(message,"Renaming %s to %s",currentName,oldName);
       logMessage(message);

       /* first remove oldName so rename will proceed w/o error;
          ignore errors */
       if ((ret = remove(oldName)) != 0) {
	  sprintf(message,"remove old file failed, %d",errno);
	  logMessage(message);
       }

       if ((ret = rename(currentName,oldName)) != 0) {
	  sprintf(message,"rename current to old failed, %d",errno);
	  logMessage(message);
       }

       /* rename blah.new to blah */
       if (rename(dent.cFileName,currentName) != 0) {
	  sprintf(message,"rename new to current failed, %d",errno);
	  logMessage(message);
       }
       
       if (FindNextFile(hFind,&dent) == 0) {
          break;
       }

     } /* while files */

     FindClose(hFind);

     logMessage("rename files done");

     return 1;
}

/* ************************************************************* */

static LRESULT CALLBACK WindowProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM  lParam )
{
    switch  ( uMsg ) {

        case WM_CREATE:
           return  0;

        case WM_COMMAND:
	   return 0;

        case WM_INITMENUPOPUP:
           return 0;

        case WM_CLOSE:
	   return 0;

        default:
           return DefWindowProc( hWnd, uMsg, wParam, lParam );
    }
}

/* ************************************************************* */

void UnregisterMainWndClass( HINSTANCE hInstance )
{
    UnregisterClass ( THIS_CLASSNAME, hInstance );
}

/* ************************************************************* */

void RegisterMainWndClass( HINSTANCE hInstance )
{
    WNDCLASSEX wclx;
    memset( &wclx, 0, sizeof( wclx ) );

    wclx.cbSize         = sizeof ( wclx );
    wclx.style          = 0;
    wclx.lpfnWndProc    = &WindowProc ;
    wclx.cbClsExtra     = 0;
    wclx.cbWndExtra     = 0;
    wclx.hInstance      = hInstance;
    //wclx.hIcon        = LoadIcon (hInstance,MAKEINTRESOURCE(IDI_TRAYICON));
    //wclx.hIconSm      = LoadSmallIcon( hInstance, IDI_TRAYICON );
    wclx.hCursor        = LoadCursor ( NULL, IDC_ARROW );
    wclx.hbrBackground  = (HBRUSH )( COLOR_BTNFACE + 1 );  //  COLOR_* + 1 is
                                                           //  special magic.
    wclx.lpszMenuName   = NULL ;
    wclx.lpszClassName  = THIS_CLASSNAME;

    RegisterClassEx ( &wclx );
}

/* ************************************************************* */

int WINAPI WinMain ( HINSTANCE hInst, HINSTANCE prev, LPSTR cmdline, int show )
{
    HWND    hWnd        = NULL;
    HWND    hPrev       = NULL;
    WSADATA wsaData;
    char messageStr[256];
    int ret;

    if (WSAStartup(0x202,&wsaData) == SOCKET_ERROR) {
        perror("WSAStartup failed");
        WSACleanup();
        exit(-1);
    }

    /* set our cwd so that we can open files w/o full pathames */
    /* set this up so that we dont have to explicitly state the path */
    if (get_home_dir("restartxmpd.exe",mydir,511) > 0) {
       _chdir(mydir);
    }
    else {
       _chdir("c:\\program files\\cartographer");
    }

    logMessage("restartxmpd.exe starting");
    logMessage(cmdline);

    if ( hPrev = FindWindow( THIS_CLASSNAME, THIS_TITLE ) )
        return 0; //SendMessage ( hPrev, WM_CLOSE, 0, 0 );

    //  We have to have a window, even though we never show it.  This is 
    //  because the tray icon uses window messages to send notifications to 
    //  its owner.  Starting with Windows 2000, you can make some kind of 
    //  "message target" window that just has a message queue and nothing
    //  much else, but we'll be backwardly compatible here.
    RegisterMainWndClass( hInst );

    hWnd = CreateWindow ( THIS_CLASSNAME, THIS_TITLE,
                         0, 0, 0, 100, 100, NULL, NULL, hInst, NULL );

    if ( ! hWnd ) {
        MessageBox ( NULL, "Ack! I can't create the window!", THIS_TITLE, 
                    MB_ICONERROR | MB_OK | MB_TOPMOST );
        logMessage("cannot create window");
        return 1;
    }

    /* stop Cartographer service */
    if ((ret = stopService(NULL,0)) < 0) {
        MessageBox ( NULL, "Cant stop the cartographer service", THIS_TITLE, 
                    MB_ICONERROR | MB_OK | MB_TOPMOST );
        sprintf(messageStr,"cannot stop service %d, trying process",ret);
        logMessage(messageStr);
	if ((ret = stopProcess(NULL,"xmpd-win32",0)) < 0) {
	   sprintf(messageStr,"cannot stop process, %d",ret);
           logMessage(messageStr);
        }
    }
    else {
       sprintf(messageStr,"stopService succeeded, %d",ret);
    }

    logMessage("Sleeping for a while");

    /* wait up to 30 seconds for Cartographer to stop gracefully */
    Sleep(35*1000);

    /* rename current files to .old, .new files to current */
    /* rename .new files to current */
    logMessage("going to rename files");
    renameFiles();

    /* restart cartographer but first wait a few more seconds for
       any sockets in TIME_WAIT to close */
    Sleep(30*1000);

    logMessage("going to restart service");

    if ((ret = startService(NULL,0)) < 0) {
        MessageBox ( NULL, "Cant start the cartographer service", THIS_TITLE, 
                    MB_ICONERROR | MB_OK | MB_TOPMOST ); 
        sprintf(messageStr,"startService failed, %d",ret);
        logMessage(messageStr);
    }
    sprintf(messageStr,"startService succeeded, %d",ret);
    logMessage(messageStr);

    /* done, double-check -- XXX */

    //MessageBox ( NULL, "Cartographer Service has been bounced", THIS_TITLE, 
    //           MB_ICONERROR | MB_OK | MB_TOPMOST );


    /* unregister our window and exit */
    UnregisterMainWndClass( hInst );

    logMessage("restartxmpd.exe done");

    if (fp != NULL) {
       fclose(fp);
    }

    return 0;
}

/* ************************************************************* */

int get_home_dir(char *myname, char *dirname, int maxlen)
{
     char pathname[1024], *cstr, *cptr;
     char program[128];

     /* just in case */
     pathname[0] = 0;          

     if (_getcwd(pathname,1023) == NULL) {
        return -1;
     }

     cstr = GetCommandLine();

     DBG2("get_home_dir: cwd:'%s' cmdline is '%s'\n",pathname,cstr);

     /* find the last backslash */
     sprintf(program,"\\%s",myname);
     cptr = strstr(cstr,program);
     if (cptr != NULL) {
        *cptr = 0; 
     }
     if (*cstr == '\"')
        cstr++;

     _snprintf(dirname,maxlen,"%s",cstr);
     dirname[maxlen-1] = 0;

     DBG1("get_home_dir: I live at '%s'\n",dirname); 

     return 1;
}

/* ************************************************************* */
