/* 
   COPYRIGHT 2006 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - setuplib.c */
/* rdk@krupczak.org */

/* NT setup related functions, etc. */

#ifdef WIN32

#include <windows.h>
#include <direct.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <netcon.h>
#include <Winnetwk.h>
#include <Psapi.h>
#include <Tlhelp32.h>

#include "setuplib.h"

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* check, add, remove share from target system */

#if 0

this function does work using NetShareGetInfo()

int checkShareOld(char *target, int verbose)
{
     PSHARE_INFO_2 sinfoPtr;
     char share[128], server[128];
     wchar_t shareWide[128];
     wchar_t serverWide[128];
     int ret;
     char aStr[256];

     if ((target == NULL) || (target[0] == 0))
        return -1;

     /* prefix target with \\ just so we dont have any problems with
        the APIs */

     sprintf(share,"C$");
     sprintf(server,"\\\\%s",target);
     mbstowcs(serverWide,server,128);
     mbstowcs(shareWide,share,128);

     if ((ret = NetShareGetInfo(serverWide,shareWide,2,
			(LPBYTE *)&sinfoPtr)) != ERROR_SUCCESS) {
       fprintf(stderr,"failed to get share info for %s on target '%s',%d,%d\n",
               share,server,ret,GetLastError());
        return -1;
     }

     fprintf(stderr,"Got share info from target '%s'\n",target);

     fprintf(stderr,"Share type %ld, permissions %ld\n",
             sinfoPtr->shi2_type,
             sinfoPtr->shi2_permissions);
 
     aStr[0] = 0;
     wcstombs(aStr,sinfoPtr->shi2_remark,255);
     fprintf(stderr,"Share remark '%s'\n",aStr);

     aStr[0] = 0;
     wcstombs(aStr,sinfoPtr->shi2_netname,255);
     fprintf(stderr,"Share netname '%s'\n",aStr);

     aStr[0] = 0;
     wcstombs(aStr,sinfoPtr->shi2_path,255);
     fprintf(stderr,"Share path '%s'\n",aStr);

     NetApiBufferFree(sinfoPtr);

     return 1;
}

#endif /* 0 */

/* ************************************************************* */
/* get the agent pathname based on where I live or defaults */

int getAgentPathname(char *agentPath, int maxlength, int verbose) 
{
     char aPath[_MAX_PATH];
     struct _stat statbuf;

     if ((agentPath == NULL) || (maxlength < 25)) {
        fprintf(stderr,"Unable to determine path\n");
        return -1;
     }

     /* check to make sure binary is in default location */
     if (_stat(PATHNAME,&statbuf) < 0) {
        if (verbose)
           fprintf(stderr,"Agent is not in default path, dealing with it\n");

        /* find out where I live; on windows x64, sometimes
           program files live in a different directory; I'm not being
           run as a service.  So, its not always safe to assume
           'c:\program files\cartographer' */

        if (_getcwd(aPath,2047) == NULL) {
           fprintf(stderr,"_getcwd failed: %d\n",GetLastError());
           return -1;
        }
 
        if (verbose)
	   fprintf(stderr,"_getcwd returned '%s'\n",aPath);

        sprintf(agentPath,"%s\\%s",aPath,AGENTNAME);
        if (_stat(agentPath,&statbuf) < 0) {
	  fprintf(stderr,"Could not properly place agent binary in '%s'\n",
                  agentPath);
          return -1;
        }
     }
     else {
        strncpy(agentPath,PATHNAME,maxlength-1);
        agentPath[maxlength-1] = 0;
     }

     return 1;
}

/* ************************************************************* */
/* check to see if we can get at C$ on target machine */

int checkShare(char *target, int verbose)
{
     NETRESOURCE nr, *nrPtr;
     u_char buffer[512];
     DWORD bufSize,ret;
     char netShare[256],*aPtr;

     bzero((void *)&nr,sizeof(nr));

     sprintf(netShare,"\\\\%s\\C$",target);

     nr.lpRemoteName = netShare;

     bufSize = 512;
     ret = WNetGetResourceInformation(&nr,buffer,&bufSize,&aPtr);

     if (ret == NO_ERROR) {
        if (verbose)
           fprintf(stderr,"Got resource information on C$ from target '%s'\n",
                  target);
        nrPtr = (NETRESOURCE *)buffer;
 
        if (verbose)
           fprintf(stderr,"Resource '%s' type %d provider '%s'\n",
                   nrPtr->lpRemoteName,
                   nrPtr->dwType,
                   nrPtr->lpProvider);
     }
     else {
        if (verbose)
           fprintf(stderr,"unable to get resource info from '%s', %d %d\n",
                   ret,GetLastError());
        return -1;
     }

     return 1;
}

/* ************************************************************* */
/* mount C$ share from target machine */

int addShare(char *target, int verbose) 
{ 
     NETRESOURCE nr;
     char netShare[256];
     DWORD ret,dwFlags;

     bzero((void *)&nr,sizeof(nr));

     sprintf(netShare,"\\\\%s\\C$",target);
     nr.lpRemoteName = netShare;

     dwFlags = 0;
     ret = WNetAddConnection2(&nr,
                              NULL,NULL, /* null password/username */
                              dwFlags);

     if (ret == NO_ERROR) {
        if (verbose)
           fprintf(stderr,"Connected resource C$ from target '%s'\n",target);
     }
     else {
        if (verbose)
           fprintf(stderr,"Failed to connect resource, %d %d\n",ret,
                   GetLastError());
     }

     return 1; 
}

/* ************************************************************* */

int removeShare(char *target, int verbose) 
{
     char netShare[256];
     DWORD ret,dwFlags;
    
     sprintf(netShare,"\\\\%s\\C$",target);

     dwFlags = 0;
     ret = WNetCancelConnection2(netShare,dwFlags,TRUE);
     if (ret == NO_ERROR) {
        if (verbose)
           fprintf(stderr,"Removed connection to resource C$ on target '%s'\n",
                   target);
     }
     else {
        if (verbose)
           fprintf(stderr,"Failed to remove resource C$ %d %d\n",ret,
                   GetLastError());
     }

     return 1; 
}

/* ************************************************************* */
/* discover share on remote target; we can do this by enum functions 
   or by trial and error; fill in the discovered share in
   shareName up to maxSize chars */

int discoverShare(char *target, int verbose, char *shareName, int maxSize)
{

     /* get remote system's systemRoot */
     /* try C$ */
     /* try C */
     /* try D$ */
     /* try D */

     return 0;
}

/* ************************************************************* */
/* copy Cartographer files from current directory to remote
   target in default locations; copy on top of existing files
   w/o errors if necessary; share must already be mounted 
*/
   
int copyFilesToShare(char *target, int verbose)
{

     /* mkdir c:\program files\cartographer XXX */

     /* copy the files themselves XXX */


     return 1;
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */

/* verify, create, check, or remove Cartographer service from
   remote system */

int verifyServiceAccess(char *target, int verbose)
{
     SC_HANDLE sch;

     /* if target is NULL, we are examining local system */

     /* open service control manager of remote machine */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM on target '%s', error %d\n",
                   target,
                   GetLastError());
        return -1;
     }
     if (verbose)
        fprintf(stderr,"Opened SCM on target '%s'\n",target);

     /* we can access service control manager on remote system */
     CloseServiceHandle(sch);

     return 1;
}

/* ************************************************************* */
/* stop/kill a running xmpd-win32 process */

int stopProcess(char *target, char *procname, int verbose)
{
     DWORD pid;
     HANDLE phandle,hProcessSnap;
     PROCESSENTRY32 pe32;

     /* right now, we dont support stopping processes on other machines */
     if (target != NULL) {
        return -1;
     }

     if (verbose)
        printf("stopProcess: looking to kill '%s'\n",procname);

     /* find the PID of the process first */

     hProcessSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS,0);
     if (hProcessSnap == INVALID_HANDLE_VALUE) {
        fprintf(stderr,"stopProcess: EnumProcs failed, %d\n",
                GetLastError());
        return -2;
     }
     pe32.dwSize = sizeof(PROCESSENTRY32);
     if (Process32First(hProcessSnap,&pe32) == FALSE) {
        CloseHandle(hProcessSnap);
        fprintf(stderr,"stopProcess: Process32First failed, %d\n",
                GetLastError());
        return -3;
     }

     pid = -1;
     do {

       if (verbose)
	  printf("Looking at process %d '%s'\n",
                 pe32.th32ProcessID,
	 	 pe32.szExeFile,procname);

       if (strstr(pe32.szExeFile,procname) != NULL) {
          pid = pe32.th32ProcessID;
          break;
       }

     } while (Process32Next(hProcessSnap,&pe32));

     CloseHandle(hProcessSnap);

     if (pid == -1) {
        if (verbose)        
	   printf("Did not find process '%s'\n",procname);
        return -4;
     }

     /* we now have pid for xmpd-win32.exe */
     if ((phandle = OpenProcess(PROCESS_ALL_ACCESS,FALSE,pid)) == NULL) {
        if (verbose)        
	   printf("Failed to open process\n");
        return -5;
     }

     if (TerminateProcess(phandle,0) == 0) {
        if (verbose) {
	   fprintf(stderr,"Unable to kill process %d\n",pid);
           CloseHandle(phandle);           
           return -6;
        }
     }
     else {
        if (verbose) {
	   fprintf(stderr,"Successfully killed process %d\n",pid);
        }
     }
     CloseHandle(phandle);

     return 1;
}

/* ************************************************************* */
/* stop service; return <0 if error, 2 if already stopped, 
   1 if stopped */

int stopService(char *target, int verbose)
{
     SC_HANDLE sch,sh;
     SERVICE_STATUS_PROCESS status;
     SERVICE_STATUS st;
     DWORD bytesNeeded,ret;

     /* open the service control manager */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        return -4;
     }

     /* open the service */
     sh = OpenService(sch,SERVICENAME,SERVICE_ALL_ACCESS);
     if (sh == NULL) {
        CloseServiceHandle(sch);
        return -3;
     }

     /* query the service status to see if its running */
     ret = QueryServiceStatusEx(sh,SC_STATUS_PROCESS_INFO,
                                (unsigned char *)&status,
                                sizeof(status),&bytesNeeded);
     if (ret == 0) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        return -2;
     }
     if (status.dwCurrentState == SERVICE_STOPPED) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        return 2;
     }

     /* now, stop the service */
     if ( ControlService(sh,SERVICE_CONTROL_STOP,&st) == FALSE) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        /* return GetLastError(); */
        return -1;
     }

     /* we should hang out and make sure its really stopped XXX */

     CloseServiceHandle(sch);
     CloseServiceHandle(sh);

     return 1; 
}

/* ************************************************************* */
/* start the service; return <0 if error, return 2 if it is
   already running; 1 if started */

int startService(char *target, int verbose)
{
     SC_HANDLE sch,sh;
     SERVICE_STATUS_PROCESS status;
     DWORD bytesNeeded,ret;

     /* open the service control manager */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        return -4;
     }

     /* open the service */
     sh = OpenService(sch,SERVICENAME,SERVICE_ALL_ACCESS);
     if (sh == NULL) {
        CloseServiceHandle(sch);
        return -3;
     }

     /* if target is NULL, we are targeting local system */

     /* query the service status to see if its running */
     ret = QueryServiceStatusEx(sh,SC_STATUS_PROCESS_INFO,
                                (unsigned char *)&status,
                                sizeof(status),&bytesNeeded);
     if (ret == 0) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        return -2;
     }
     if (status.dwCurrentState != SERVICE_STOPPED) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        return 2;
     }

     /* start the service */
     if ((ret = StartService(sh, 0, NULL)) == 0) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        /* return GetLastError(); errors are positive ints so we cant
           return error number */
        return -1;
     }

     /* we should hang out and make sure its really started XXX */
 
     /* close everything */
     CloseServiceHandle(sch);
     CloseServiceHandle(sh);

     return 1; 
}

/* ************************************************************* */
/* use CreateService() to create service with SC Manager; this
   function also creates the registry keys
*/

int createService(char *target, int verbose) 
{ 
     SC_HANDLE sch,sh;
     SERVICE_DESCRIPTION serviceDescription;
     char *descr = SERVICEDESCRIPTION;
     char agentPath[_MAX_PATH];

     /* figure out the agent's path */
     if (getAgentPathname(agentPath,_MAX_PATH-1,verbose) < 0) {
        fprintf(stderr,"Unable to determine agent location\n");
        return -1;
     }

     if (verbose) {
        fprintf(stderr,"Creating service for agent '%s'\n",agentPath);
     }

     /* if target is NULL, we are examining local system */

     /* open SC Manager and get handle */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM, %d\n",GetLastError());
        return -1;
     }

     /* make the call */
     sh = CreateService(sch,
                        SERVICENAME,
                        SERVICEDISPLAYNAME,
                        SC_MANAGER_ALL_ACCESS,
                        SERVICE_WIN32_OWN_PROCESS,
                        SERVICE_AUTO_START,
                        SERVICE_ERROR_IGNORE,
                        agentPath,
                        NULL, /* order group */
                        NULL, /* tagID */
                        NULL, /* dependencies */
                        NULL, /* account name; local system default */
                        NULL); /* password */

     if (sh == NULL) {
        if (verbose)
           fprintf(stderr,"failed to create service, %d\n",GetLastError());
        CloseServiceHandle(sch);
        return -1;
     }

     /* add description */
     serviceDescription.lpDescription = descr;
     ChangeServiceConfig2(sh,SERVICE_CONFIG_DESCRIPTION,
                          (void *)&serviceDescription);

     /* close handles */
     CloseServiceHandle(sch);
     CloseServiceHandle(sh);

     if (verbose)
        fprintf(stderr,"created service, verifying\n");

     /* verify that its installed */
     return verifyService(target,verbose);
}

/* ************************************************************* */

/* check to see if Cartographer service is installed; -1 if failure,
   0 if not installed, 1 if installed; we do not determine if
   the service is currently running  */

int checkService(char *target, int verbose)
{ 
     SC_HANDLE sch, sh;
     int ret = 0;

     /* open service control manager of remote machine */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM on target '%s', error %d\n",
                   target,
                   GetLastError());
        return -1;
     }
 
     if (verbose)
        fprintf(stderr,"Opened SCM on target '%s'\n",target);

     /* open our service on remote machine */
     sh = OpenService(sch,SERVICENAME,SERVICE_QUERY_CONFIG);
     if (sh == NULL) {
        if (verbose)
           fprintf(stderr,"service '%s' is not installed on target '%s'\n",
                   SERVICENAME,target);
        ret = 0; /* not installed */
     }
     else {
       if (verbose)
          fprintf(stderr,"Service '%s' is installed on target '%s'\n",
                  SERVICENAME,target);
        CloseServiceHandle(sh);
        ret = 1;
     }

     CloseServiceHandle(sch);

     return ret;
}

/* ************************************************************* */

int removeService(char *target, int verbose) 
{ 
     /* if target is NULL, we are examining local system */
     SC_HANDLE sch,sh;

     /* open SC Manager and get handle */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM, %d\n",GetLastError());
        return -1;
     }

     /* open our service */
     sh = OpenService(sch,SERVICENAME,SERVICE_QUERY_CONFIG|DELETE);
     if (sh == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open service, %d\n",GetLastError());
        CloseServiceHandle(sch);
        return -1;
     }

     if (DeleteService(sh) == 0) {
        if (verbose)
           fprintf(stderr,"failed to delete service, %d\n",
                GetLastError());
     }
     else {
        if (verbose)
           fprintf(stderr,"Service delete\n");
     }

     CloseServiceHandle(sch);
     CloseServiceHandle(sh);

     return 1;     
}

/* ************************************************************* */

int listServices(char *target, int verbose)
{
     SC_HANDLE sch;
     ENUM_SERVICE_STATUS *ess;
     u_char *buf = NULL;
     int bufsize = 0;
     int needed,ret,num,resume,i;

     /* if target is NULL, we are examining local system */

     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM, %d\n",GetLastError());
        return -1;
     }

     if (verbose)
        fprintf(stderr,"opened SCM\n");

     resume = 0;
     while (1) {

       ret = EnumServicesStatus(sch,SERVICE_WIN32,SERVICE_STATE_ALL,
                               (ENUM_SERVICE_STATUS *)buf,bufsize,
                                &needed,&num,&resume);
       if (verbose)
          fprintf(stderr,"Initial EnumServicesStatus returned %d\n",ret);
        
       if (ret == 0) {
          if (GetLastError() == ERROR_MORE_DATA) {
  	     bufsize = needed;
	     if ((buf = realloc(buf,needed)) == NULL) {
	         if (verbose)
	            fprintf(stderr,"realloc of service structs failed, %d\n",
                           GetLastError());    
                 CloseServiceHandle(sch);
                 return -1;
	      }
          }
          else {
	     break; /* from while loop */
          }
       }
       else {
          if (verbose)
             fprintf(stderr,"EnumServiceStatus returned %d services\n",num);
          if (num == 0)
	     break;
          ess = (ENUM_SERVICE_STATUS *)buf;
          for (i=0; i<num; i++) {
	      if (verbose)
                 fprintf(stderr,"'%s', '%s', Type:%d\n",
   	                 ess[i].lpServiceName,
                         ess[i].lpDisplayName,
                         ess[i].ServiceStatus.dwServiceType
                        );
          }
          /* if we've listed a bunch but needed is 0, we can break */
          if (needed == 0)
	     break;
       }


     } /* while */

     CloseServiceHandle(sch);
     return 1;
}

/* ************************************************************* */

int verifyService(char *target, int verbose)
{
     SC_HANDLE sch,sh;
     char buf[8096];
     int bytesNeeded;
     QUERY_SERVICE_CONFIG *q;

     /* if target is NULL, we are examining local system */

     /* open service control manager again */
     sch = OpenSCManager(target,SERVICES_ACTIVE_DATABASE,
                         SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open SCM, %d\n",GetLastError());
        return -1;
     }

     /* open our service */
     sh = OpenService(sch,SERVICENAME,SERVICE_QUERY_CONFIG);
     if (sh == NULL) {
        if (verbose)
           fprintf(stderr,"failed to open service, %d\n",GetLastError());
        CloseServiceHandle(sch);
        return -1;
     }

     /* QueryServiceConfig; dont worry about needed and buf size 
        since we are querying ourselves only */
     if (QueryServiceConfig(sh,(QUERY_SERVICE_CONFIG *)buf,8096,
                            &bytesNeeded) == 0) {
        if (verbose)
           fprintf(stderr,"failed to query service, %d\n",GetLastError());
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        return -1;  
     }

     q = (QUERY_SERVICE_CONFIG *)buf;

     if (verbose) {
        printf("Service: %s\n",SERVICENAME);
        printf("Path: %s\n",q->lpBinaryPathName);
        printf("DisplayName: %s\n",q->lpDisplayName);
     }

     CloseServiceHandle(sh);
     CloseServiceHandle(sch);

     return 1;
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* registry related setup functions */

/* 
   HKLM\\Software\\Krupczak.org\\Cartographer
      Install Directory
      Path
      ProductName
      LicenseInfo
      SupportInfo
*/

int verifyRegistryAccess(char *target, int verbose)
{
     HKEY subKey,subSubKey;
     char systemRoot[256];
     int systemRootSize;
     DWORD rtype;

     /* open registry of remote machine */
     if (RegConnectRegistry(target,HKEY_LOCAL_MACHINE,&subKey)  
         != ERROR_SUCCESS) {
        if (verbose)
           fprintf(stderr,"failed to open registry on target '%s', %d\n",
                   target,GetLastError());
        return -1;
     }
     if (verbose)
        fprintf(stderr,"Opened registry on target '%s'\n",target);

     if (RegOpenKeyEx(subKey,
	             "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",
		      0,KEY_READ,&subSubKey) != ERROR_SUCCESS) {
         if (verbose)
            fprintf(stderr,"failed to access sub key on target '%s', %d\n",
                   target,GetLastError());
     }

     /* get systemroot of remote machine */
     /* \\HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\SystemRoot */ 
     systemRootSize = sizeof(systemRoot);
     if (RegQueryValueEx(subSubKey,
	           "SystemRoot",
                    NULL,
		    &rtype,
                    systemRoot,
	   	    &systemRootSize) != ERROR_SUCCESS) {
        if (verbose)
           fprintf(stderr,"failed to get target '%s' systemRoot, %d\n",
                   target,
                   GetLastError());
     }
     else {
        if (verbose)
           fprintf(stderr,"Target '%s' has systemRoot '%s'\n",target,
                   systemRoot);
     }

     RegCloseKey(subSubKey);
     RegCloseKey(subKey);

     return 1;
}

/* ************************************************************* */

int createRegistryEntries(char *target, int verbose)
{ 
     return 0;
}

/* ************************************************************* */
/* check registry for Cartographer */

int checkRegistryEntries(char *target, int verbose)
{ 
     return 1; 
}

/* ************************************************************* */
/* remove entries from registry (either local or remote) */

int removeRegistryEntries(char *target, int verbose) { return -1; }


/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */

#endif /* WIN32 */
